package org.apache.commons.collections4;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import org.junit.Test;

/** 
 * Abstract test class for {@link java.lang.Object} methods and contracts.
 * <p>
 * To use, simply extend this class, and implement
 * the {@link #makeObject()} method.
 * <p>
 * If your {@link Object} fails one of these tests by design,
 * you may still use this base set of cases.  Simply override the
 * test case (method) your {@link Object} fails.
 * 
 * @version $Id$
 */
public abstract class AbstractObjectTest extends BulkTest {
    /** 
     * Current major release for Collections
     */
public static final int COLLECTIONS_MAJOR_VERSION = 4;

    /** 
     * JUnit constructor.
     * 
     * @param testName  the test class name
     */
public AbstractObjectTest(final String testName) {
        super(testName);
    }

    /** 
     * Implement this method to return the object to test.
     * 
     * @return the object to test
     */
public abstract Object makeObject();

    /** 
     * Override this method if a subclass is testing an object
     * that cannot serialize an "empty" Collection.
     * (e.g. Comparators have no contents)
     * 
     * @return true
     */
public boolean supportsEmptyCollections() {
        return true;
    }

    /** 
     * Override this method if a subclass is testing an object
     * that cannot serialize a "full" Collection.
     * (e.g. Comparators have no contents)
     * 
     * @return true
     */
public boolean supportsFullCollections() {
        return true;
    }

    /** 
     * Is serialization testing supported.
     * Default is true.
     */
public boolean isTestSerialization() {
        return true;
    }

    /** 
     * Returns true to indicate that the collection supports equals() comparisons.
     * This implementation returns true;
     */
public boolean isEqualsCheckable() {
        return true;
    }

    public void testObjectEqualsSelf() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testObjectEqualsSelf");
            final Object obj = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),998,obj);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),997,obj);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),997,obj);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    public void testEqualsNull() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testEqualsNull");
            final Object obj = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),996,obj);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),995,obj.equals(null));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    public void testObjectHashCodeEqualsSelfHashCode() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testObjectHashCodeEqualsSelfHashCode");
            final Object obj = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1006,obj);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1005,obj.hashCode());
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1005,obj.hashCode());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    public void testObjectHashCodeEqualsContract() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testObjectHashCodeEqualsContract");
            final Object obj1 = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1003,obj1);
            if (obj1.equals(obj1)) {
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),999,obj1.hashCode());
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),999,obj1.hashCode());
            } 
            final Object obj2 = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1004,obj2);
            if (obj1.equals(obj2)) {
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1000,obj1.hashCode());
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1001,obj2.hashCode());
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1002,obj2.equals(obj1));
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    protected Object serializeDeserialize(final Object obj) throws Exception {
        final ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),977,buffer);
        final ObjectOutputStream out = new ObjectOutputStream(buffer);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),978,out);
        out.writeObject(obj);
        out.close();
        final ObjectInputStream in = new ObjectInputStream(new ByteArrayInputStream(buffer.toByteArray()));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),979,in);
        final Object dest = in.readObject();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),980,dest);
        in.close();
        return dest;
    }

    public void testSerializeDeserializeThenCompare() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testSerializeDeserializeThenCompare");
            final Object obj = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1009,obj);
            if ((obj instanceof Serializable) && (isTestSerialization())) {
                final Object dest = serializeDeserialize(obj);
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1010,dest);
                if (isEqualsCheckable()) {
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1007,obj);
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1008,dest);
                } 
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Sanity check method, makes sure that any Serializable
     * class can be serialized and de-serialized in memory,
     * using the handy makeObject() method
     * 
     * @throws IOException
     * @throws ClassNotFoundException
     */
@Test(timeout = 1000)
    public void testSimpleSerialization() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testSimpleSerialization");
            final Object o = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1011,o);
            if ((o instanceof Serializable) && (isTestSerialization())) {
                final byte[] objekt = writeExternalFormToBytes(((Serializable)(o)));
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1012,objekt);
                readExternalFormFromBytes(objekt);
                readExternalFormFromBytes(objekt);
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Sanity check method, makes sure that any Serializable
     * class can be serialized and de-serialized in memory,
     * using the handy makeObject() method
     * 
     * @throws IOException
     * @throws ClassNotFoundException
     */
@Test(timeout = 1000)
    public void testSimpleSerialization_remove369() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testSimpleSerialization_remove369");
            final Object o = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1011,o);
            if ((o instanceof Serializable) && (isTestSerialization())) {
                final byte[] objekt = writeExternalFormToBytes(((Serializable)(o)));
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1012,objekt);
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Tests serialization by comparing against a previously stored version in SVN.
     * If the test object is serializable, confirm that a canonical form exists.
     */
public void testCanonicalEmptyCollectionExists() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testCanonicalEmptyCollectionExists");
            if (((supportsEmptyCollections()) && (isTestSerialization())) && (!(skipSerializedCanonicalTests()))) {
                final Object object = makeObject();
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),989,object);
                if (object instanceof Serializable) {
                    final String name = getCanonicalEmptyCollectionName(object);
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),990,name);
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),987,(("Canonical empty collection (" + name) + ") is not in SVN"));
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),988,new java.io.File(name).exists());
                } 
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Tests serialization by comparing against a previously stored version in SVN.
     * If the test object is serializable, confirm that a canonical form exists.
     */
public void testCanonicalFullCollectionExists() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testCanonicalFullCollectionExists");
            if (((supportsFullCollections()) && (isTestSerialization())) && (!(skipSerializedCanonicalTests()))) {
                final Object object = makeObject();
                fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),993,object);
                if (object instanceof Serializable) {
                    final String name = getCanonicalFullCollectionName(object);
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),994,name);
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),991,(("Canonical full collection (" + name) + ") is not in SVN"));
                    fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),992,new java.io.File(name).exists());
                } 
            } 
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Get the version of Collections that this object tries to
     * maintain serialization compatibility with. Defaults to 4, due to
     * the package change to collections4 introduced in version 4.
     * 
     * This constant makes it possible for TestMap (and other subclasses,
     * if necessary) to automatically check SVN for a versionX copy of a
     * Serialized object, so we can make sure that compatibility is maintained.
     * See, for example, TestMap.getCanonicalFullMapName(Map map).
     * Subclasses can override this variable, indicating compatibility
     * with earlier Collections versions.
     * 
     * @return The version, or <code>null</code> if this object shouldn't be
     * tested for compatibility with previous versions.
     */
public String getCompatibilityVersion() {
        return "4";
    }

    protected String getCanonicalEmptyCollectionName(final Object object) {
        final StringBuilder retval = new StringBuilder();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),982,retval);
        retval.append(BulkTest.TEST_DATA_PATH);
        String colName = object.getClass().getName();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),983,colName);
        colName = colName.substring(((colName.lastIndexOf(".")) + 1), colName.length());
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),981,colName);
        retval.append(colName);
        retval.append(".emptyCollection.version");
        retval.append(getCompatibilityVersion());
        retval.append(".obj");
        return retval.toString();
    }

    protected String getCanonicalFullCollectionName(final Object object) {
        final StringBuilder retval = new StringBuilder();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),985,retval);
        retval.append(BulkTest.TEST_DATA_PATH);
        String colName = object.getClass().getName();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),986,colName);
        colName = colName.substring(((colName.lastIndexOf(".")) + 1), colName.length());
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),984,colName);
        retval.append(colName);
        retval.append(".fullCollection.version");
        retval.append(getCompatibilityVersion());
        retval.append(".obj");
        return retval.toString();
    }

    /** 
     * Write a Serializable or Externalizable object as
     * a file at the given path.  NOT USEFUL as part
     * of a unit test; this is just a utility method
     * for creating disk-based objects in SVN that can become
     * the basis for compatibility tests using
     * readExternalFormFromDisk(String path)
     * 
     * @param o Object to serialize
     * @param path path to write the serialized Object
     * @exception IOException
     */
protected void writeExternalFormToDisk(final Serializable o, final String path) throws IOException {
        final FileOutputStream fileStream = new FileOutputStream(path);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1013,fileStream);
        writeExternalFormToStream(o, fileStream);
    }

    /** 
     * Converts a Serializable or Externalizable object to
     * bytes.  Useful for in-memory tests of serialization
     * 
     * @param o Object to convert to bytes
     * @return serialized form of the Object
     * @exception IOException
     */
protected byte[] writeExternalFormToBytes(final Serializable o) throws IOException {
        final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),973,byteStream);
        writeExternalFormToStream(o, byteStream);
        return byteStream.toByteArray();
    }

    /** 
     * Reads a Serialized or Externalized Object from disk.
     * Useful for creating compatibility tests between
     * different SVN versions of the same class
     * 
     * @param path path to the serialized Object
     * @return the Object at the given path
     * @exception IOException
     * @exception ClassNotFoundException
     */
protected Object readExternalFormFromDisk(final String path) throws IOException, ClassNotFoundException {
        final FileInputStream stream = new FileInputStream(path);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),975,stream);
        return readExternalFormFromStream(stream);
    }

    /** 
     * Read a Serialized or Externalized Object from bytes.
     * Useful for verifying serialization in memory.
     * 
     * @param b byte array containing a serialized Object
     * @return Object contained in the bytes
     * @exception IOException
     * @exception ClassNotFoundException
     */
protected Object readExternalFormFromBytes(final byte[] b) throws IOException, ClassNotFoundException {
        final ByteArrayInputStream stream = new ByteArrayInputStream(b);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),974,stream);
        return readExternalFormFromStream(stream);
    }

    protected boolean skipSerializedCanonicalTests() {
        return java.lang.Boolean.getBoolean("org.apache.commons.collections:with-clover");
    }

    private Object readExternalFormFromStream(final InputStream stream) throws IOException, ClassNotFoundException {
        final ObjectInputStream oStream = new ObjectInputStream(stream);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),976,oStream);
        return oStream.readObject();
    }

    private void writeExternalFormToStream(final Serializable o, final OutputStream stream) throws IOException {
        final ObjectOutputStream oStream = new ObjectOutputStream(stream);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),1014,oStream);
        oStream.writeObject(o);
    }
}

