package org.apache.commons.collections4.iterators;

import org.apache.commons.collections4.AbstractObjectTest;
import java.util.Iterator;
import java.util.NoSuchElementException;
import org.junit.Test;

/** 
 * Abstract class for testing the Iterator interface.
 * <p>
 * This class provides a framework for testing an implementation of Iterator.
 * Concrete subclasses must provide the iterator to be tested.
 * They must also specify certain details of how the iterator operates by
 * overriding the supportsXxx() methods if necessary.
 * 
 * @since 3.0
 * @version $Id$
 */
public abstract class AbstractIteratorTest<E> extends AbstractObjectTest {
    /** 
     * JUnit constructor.
     * 
     * @param testName  the test class name
     */
public AbstractIteratorTest(final String testName) {
        super(testName);
    }

    /** 
     * Implement this method to return an iterator over an empty collection.
     * 
     * @return an empty iterator
     */
public abstract Iterator<E> makeEmptyIterator();

    /** 
     * Implements the abstract superclass method to return the full iterator.
     * 
     * @return a full iterator
     */
@Override
    public abstract Iterator<E> makeObject();

    /** 
     * Whether or not we are testing an iterator that can be empty.
     * Default is true.
     * 
     * @return true if Iterator can be empty
     */
public boolean supportsEmptyIterator() {
        return true;
    }

    /** 
     * Whether or not we are testing an iterator that can contain elements.
     * Default is true.
     * 
     * @return true if Iterator can be full
     */
public boolean supportsFullIterator() {
        return true;
    }

    /** 
     * Whether or not we are testing an iterator that supports remove().
     * Default is true.
     * 
     * @return true if Iterator supports remove
     */
public boolean supportsRemove() {
        return true;
    }

    /** 
     * Allows subclasses to add complex cross verification
     */
public void verify() {
    }

    /** 
     * Test the empty iterator.
     */
@Test(timeout = 1000)
    public void testEmptyIterator() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testEmptyIterator");
            if (!(supportsEmptyIterator())) {
                return ;
            } 
            final Iterator<E> it = makeEmptyIterator();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),451,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),449,it.hasNext());
            try {
                it.next();
                it.next();
            } catch (final NoSuchElementException e) {
            }
            verify();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),450,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test the empty iterator.
     */
@Test(timeout = 1000)
    public void testEmptyIterator_add185() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testEmptyIterator_add185");
            if (!(supportsEmptyIterator())) {
                return ;
            } 
            final Iterator<E> it = makeEmptyIterator();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),451,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),449,it.hasNext());
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            verify();
            verify();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),450,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test the empty iterator.
     */
@Test(timeout = 1000)
    public void testEmptyIterator_remove150() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testEmptyIterator_remove150");
            if (!(supportsEmptyIterator())) {
                return ;
            } 
            final Iterator<E> it = makeEmptyIterator();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),451,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),449,it.hasNext());
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),450,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test normal iteration behaviour.
     */
@Test(timeout = 1000)
    public void testFullIterator() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testFullIterator");
            if (!(supportsFullIterator())) {
                return ;
            } 
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),454,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),452,it.hasNext());
            try {
                it.next();
                it.next();
            } catch (final NoSuchElementException e) {
            }
            while (it.hasNext()) {
                it.next();
                verify();
            }
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),453,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test normal iteration behaviour.
     */
@Test(timeout = 1000)
    public void testFullIterator_add187() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testFullIterator_add187");
            if (!(supportsFullIterator())) {
                return ;
            } 
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),454,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),452,it.hasNext());
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            while (it.hasNext()) {
                it.next();
                it.next();
                verify();
            }
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),453,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test normal iteration behaviour.
     */
@Test(timeout = 1000)
    public void testFullIterator_add188() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testFullIterator_add188");
            if (!(supportsFullIterator())) {
                return ;
            } 
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),454,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),452,it.hasNext());
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            while (it.hasNext()) {
                it.next();
                verify();
                verify();
            }
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),453,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test normal iteration behaviour.
     */
@Test(timeout = 1000)
    public void testFullIterator_add189() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testFullIterator_add189");
            if (!(supportsFullIterator())) {
                return ;
            } 
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),454,it);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),452,it.hasNext());
            try {
                it.next();
            } catch (final NoSuchElementException e) {
            }
            while (it.hasNext()) {
                it.next();
                verify();
            }
            try {
                it.next();
                it.next();
            } catch (final NoSuchElementException e) {
            }
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),453,it.toString());
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_add191() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_add191");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_add192() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_add192");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            verify();
            it.next();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_add193() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_add193");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            it.next();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_add194() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_add194");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            it.remove();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_add195() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_add195");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            it.remove();
            try {
                it.remove();
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_remove151() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_remove151");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            it.next();
            it.remove();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    /** 
     * Test remove behaviour.
     */
@Test(timeout = 1000)
    public void testRemove_remove152() {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testRemove_remove152");
            final Iterator<E> it = makeObject();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),455,it);
            if (!(supportsRemove())) {
                try {
                    it.remove();
                } catch (final UnsupportedOperationException ex) {
                }
                return ;
            } 
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
            verify();
            it.next();
            try {
                it.remove();
            } catch (final IllegalStateException ex) {
            }
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }
}

