package org.eclipse.jgit.api;

import org.eclipse.jgit.lib.AnyObjectId;
import java.util.ArrayList;
import java.io.ByteArrayOutputStream;
import org.eclipse.jgit.api.errors.CheckoutConflictException;
import java.util.Collection;
import org.eclipse.jgit.api.errors.ConcurrentRefUpdateException;
import org.eclipse.jgit.diff.DiffFormatter;
import org.eclipse.jgit.dircache.DirCache;
import org.eclipse.jgit.dircache.DirCacheCheckout;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import org.eclipse.jgit.api.errors.GitAPIException;
import java.util.HashMap;
import java.io.IOException;
import org.eclipse.jgit.api.errors.InvalidRebaseStepException;
import org.eclipse.jgit.api.errors.InvalidRefNameException;
import java.util.Iterator;
import org.eclipse.jgit.api.errors.JGitInternalException;
import org.eclipse.jgit.internal.JGitText;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import org.eclipse.jgit.merge.MergeStrategy;
import org.eclipse.jgit.api.errors.NoHeadException;
import org.eclipse.jgit.api.errors.NoMessageException;
import org.eclipse.jgit.lib.NullProgressMonitor;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectReader;
import java.util.regex.Pattern;
import org.eclipse.jgit.lib.PersonIdent;
import org.eclipse.jgit.lib.ProgressMonitor;
import org.eclipse.jgit.lib.RebaseTodoLine;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.api.errors.RefAlreadyExistsException;
import org.eclipse.jgit.api.errors.RefNotFoundException;
import org.eclipse.jgit.lib.RefUpdate;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.lib.RepositoryState;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevWalk;
import org.eclipse.jgit.errors.RevisionSyntaxException;
import org.eclipse.jgit.api.errors.StashApplyFailureException;
import org.eclipse.jgit.treewalk.TreeWalk;
import org.eclipse.jgit.api.errors.UnmergedPathsException;
import org.eclipse.jgit.api.errors.WrongRepositoryStateException;

/** 
 * A class used to execute a {@code Rebase} command. It has setters for all
 * supported options and arguments of this command and a {@link #call()} method
 * to finally execute the command. Each instance of this class should only be
 * used for one invocation of the command (means: one call to {@link #call()})
 * <p>
 * 
 * @see <a
 * href="http://www.kernel.org/pub/software/scm/git/docs/git-rebase.html"
 * >Git documentation about Rebase</a>
 */
public class RebaseCommand extends GitCommand<org.eclipse.jgit.api.RebaseResult> {
    /** 
     * * The name of the "rebase-merge" folder for interactive rebases.
     */
public static final String REBASE_MERGE = "rebase-merge";

    /** 
     * * The name of the "rebase-apply" folder for non-interactive rebases.
     */
private static final String REBASE_APPLY = "rebase-apply";

    /** 
     * * The name of the "stopped-sha" file
     */
public static final String STOPPED_SHA = "stopped-sha";

    private static final String AUTHOR_SCRIPT = "author-script";

    private static final String DONE = "done";

    private static final String GIT_AUTHOR_DATE = "GIT_AUTHOR_DATE";

    private static final String GIT_AUTHOR_EMAIL = "GIT_AUTHOR_EMAIL";

    private static final String GIT_AUTHOR_NAME = "GIT_AUTHOR_NAME";

    private static final String GIT_REBASE_TODO = "git-rebase-todo";

    private static final String HEAD_NAME = "head-name";

    private static final String INTERACTIVE = "interactive";

    private static final String QUIET = "quiet";

    private static final String MESSAGE = "message";

    private static final String ONTO = "onto";

    private static final String ONTO_NAME = "onto-name";

    private static final String PATCH = "patch";

    private static final String REBASE_HEAD = "head";

    private static final String AMEND = "amend";

    private static final String MESSAGE_FIXUP = "message-fixup";

    private static final String MESSAGE_SQUASH = "message-squash";

    private static final String AUTOSTASH = "autostash";

    private static final String AUTOSTASH_MSG = "On {0}: autostash";

    /** 
     * * The folder containing the hashes of (potentially) rewritten commits when
     * * --preserve-merges is used.
     */
private static final String REWRITTEN = "rewritten";

    /** 
     * * File containing the current commit(s) to cherry pick when --preserve-merges
     * * is used.
     */
private static final String CURRENT_COMMIT = "current-commit";

    private static final String REFLOG_PREFIX = "rebase:";

    /** 
     * * The available operations
     */
public enum Operation {
BEGIN, CONTINUE, SKIP, ABORT, PROCESS_STEPS;    }

    private Operation operation = Operation.BEGIN;

    private RevCommit upstreamCommit;

    private String upstreamCommitName;

    private ProgressMonitor monitor = NullProgressMonitor.INSTANCE;

    private final RevWalk walk;

    private final RebaseState rebaseState;

    private InteractiveHandler interactiveHandler;

    private boolean stopAfterInitialization = false;

    private RevCommit newHead;

    private boolean lastStepWasForward;

    private MergeStrategy strategy = MergeStrategy.RECURSIVE;

    private boolean preserveMerges = false;

    /** 
     * * @param repo
     */
protected RebaseCommand(Repository repo) {
        super(repo);
        walk = new RevWalk(repo);
        rebaseState = new RebaseState(repo.getDirectory());
    }

    /** 
     * * Executes the {@code Rebase} command with all the options and parameters
     * * collected by the setter methods of this class. Each instance of this
     * * class should only be used for one invocation of the command. Don't call
     * * this method twice on an instance.
     * *
     * * @return an object describing the result of this command
     * * @throws GitAPIException
     * * @throws WrongRepositoryStateException
     * * @throws NoHeadException
     * * @throws RefNotFoundException
     */
public RebaseResult call() throws GitAPIException, NoHeadException, RefNotFoundException, WrongRepositoryStateException {
        newHead = null;
        lastStepWasForward = false;
        checkCallable();
        checkParameters();
        try {
            switch (operation) {
                case ABORT :
                    try {
                        return abort(RebaseResult.ABORTED_RESULT);
                    } catch (IOException ioe) {
                        throw new JGitInternalException(ioe.getMessage() , ioe);
                    }
                case PROCESS_STEPS :
                case SKIP :
                case CONTINUE :
                    String upstreamCommitId = rebaseState.readFile(ONTO);
                    try {
                        upstreamCommitName = rebaseState.readFile(ONTO_NAME);
                    } catch (FileNotFoundException e) {
                        upstreamCommitName = upstreamCommitId;
                    }
                    this.upstreamCommit = walk.parseCommit(repo.resolve(upstreamCommitId));
                    preserveMerges = rebaseState.getRewrittenDir().exists();
                    break;
                case BEGIN :
                    autoStash();
                    if ((stopAfterInitialization) || (!(walk.isMergedInto(walk.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD)), upstreamCommit)))) {
                        Status status = Git.wrap(repo).status().setIgnoreSubmodules(org.eclipse.jgit.submodule.SubmoduleWalk.IgnoreSubmoduleMode.ALL).call();
                        if (status.hasUncommittedChanges()) {
                            List<java.lang.String> list = new ArrayList<java.lang.String>();
                            list.addAll(status.getUncommittedChanges());
                            return RebaseResult.uncommittedChanges(list);
                        } 
                    } 
                    RebaseResult res = initFilesAndRewind();
                    if (stopAfterInitialization)
                        return RebaseResult.INTERACTIVE_PREPARED_RESULT;
                    
                    if (res != null) {
                        autoStashApply();
                        if (rebaseState.getDir().exists())
                            org.eclipse.jgit.util.FileUtils.delete(rebaseState.getDir(), org.eclipse.jgit.util.FileUtils.RECURSIVE);
                        
                        return res;
                    } 
            }
            if (monitor.isCancelled())
                return abort(RebaseResult.ABORTED_RESULT);
            
            if ((operation) == (Operation.CONTINUE)) {
                newHead = continueRebase();
                List<org.eclipse.jgit.lib.RebaseTodoLine> doneLines = repo.readRebaseTodo(rebaseState.getPath(DONE), true);
                RebaseTodoLine step = doneLines.get(((doneLines.size()) - 1));
                if (((newHead) != null) && ((step.getAction()) != (RebaseTodoLine.Action.PICK))) {
                    RebaseTodoLine newStep = new RebaseTodoLine(step.getAction() , org.eclipse.jgit.lib.AbbreviatedObjectId.fromObjectId(newHead) , step.getShortMessage());
                    RebaseResult result = processStep(newStep, false);
                    if (result != null)
                        return result;
                    
                } 
                File amendFile = rebaseState.getFile(AMEND);
                boolean amendExists = amendFile.exists();
                if (amendExists) {
                    org.eclipse.jgit.util.FileUtils.delete(amendFile);
                } 
                if (((newHead) == null) && (!amendExists)) {
                    return RebaseResult.NOTHING_TO_COMMIT_RESULT;
                } 
            } 
            if ((operation) == (Operation.SKIP))
                newHead = checkoutCurrentHead();
            
            List<org.eclipse.jgit.lib.RebaseTodoLine> steps = repo.readRebaseTodo(rebaseState.getPath(GIT_REBASE_TODO), false);
            if ((steps.size()) == 0) {
                return finishRebase(walk.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD)), false);
            } 
            if (isInteractive()) {
                interactiveHandler.prepareSteps(steps);
                repo.writeRebaseTodoFile(rebaseState.getPath(GIT_REBASE_TODO), steps, false);
            } 
            checkSteps(steps);
            for (int i = 0 ; i < (steps.size()) ; i++) {
                RebaseTodoLine step = steps.get(i);
                popSteps(1);
                RebaseResult result = processStep(step, true);
                if (result != null) {
                    return result;
                } 
            }
            return finishRebase(newHead, lastStepWasForward);
        } catch (CheckoutConflictException cce) {
            return RebaseResult.conflicts(cce.getConflictingPaths());
        } catch (IOException ioe) {
            throw new JGitInternalException(ioe.getMessage() , ioe);
        }
    }

    private void autoStash() throws IOException, GitAPIException {
        if (repo.getConfig().getBoolean(org.eclipse.jgit.lib.ConfigConstants.CONFIG_REBASE_SECTION, org.eclipse.jgit.lib.ConfigConstants.CONFIG_KEY_AUTOSTASH, false)) {
            String message = java.text.MessageFormat.format(AUTOSTASH_MSG, Repository.shortenRefName(RebaseCommand.getHeadName(getHead())));
            RevCommit stashCommit = Git.wrap(repo).stashCreate().setRef(null).setWorkingDirectoryMessage(message).call();
            if (stashCommit != null) {
                org.eclipse.jgit.util.FileUtils.mkdir(rebaseState.getDir());
                rebaseState.createFile(AUTOSTASH, stashCommit.getName());
            } 
        } 
    }

    private boolean autoStashApply() throws IOException, GitAPIException {
        boolean conflicts = false;
        if (rebaseState.getFile(AUTOSTASH).exists()) {
            String stash = rebaseState.readFile(AUTOSTASH);
            try {
                Git.wrap(repo).stashApply().setStashRef(stash).ignoreRepositoryState(true).setStrategy(strategy).call();
            } catch (StashApplyFailureException e) {
                conflicts = true;
                RevWalk rw = new RevWalk(repo);
                ObjectId stashId = repo.resolve(stash);
                RevCommit commit = rw.parseCommit(stashId);
                updateStashRef(commit, commit.getAuthorIdent(), commit.getShortMessage());
            }
        } 
        return conflicts;
    }

    private void updateStashRef(ObjectId commitId, PersonIdent refLogIdent, String refLogMessage) throws IOException {
        Ref currentRef = repo.getRef(org.eclipse.jgit.lib.Constants.R_STASH);
        RefUpdate refUpdate = repo.updateRef(org.eclipse.jgit.lib.Constants.R_STASH);
        refUpdate.setNewObjectId(commitId);
        refUpdate.setRefLogIdent(refLogIdent);
        refUpdate.setRefLogMessage(refLogMessage, false);
        if (currentRef != null)
            refUpdate.setExpectedOldObjectId(currentRef.getObjectId());
        else
            refUpdate.setExpectedOldObjectId(ObjectId.zeroId());
        
        refUpdate.forceUpdate();
    }

    private RebaseResult processStep(RebaseTodoLine step, boolean shouldPick) throws IOException, GitAPIException {
        if (RebaseTodoLine.Action.COMMENT.equals(step.getAction()))
            return null;
        
        if (((preserveMerges) && shouldPick) && ((RebaseTodoLine.Action.EDIT.equals(step.getAction())) || (RebaseTodoLine.Action.PICK.equals(step.getAction())))) {
            writeRewrittenHashes();
        } 
        ObjectReader or = repo.newObjectReader();
        Collection<org.eclipse.jgit.lib.ObjectId> ids = or.resolve(step.getCommit());
        if ((ids.size()) != 1)
            throw new JGitInternalException("Could not resolve uniquely the abbreviated object ID");
        
        RevCommit commitToPick = walk.parseCommit(ids.iterator().next());
        if (shouldPick) {
            if (monitor.isCancelled())
                return RebaseResult.result(RebaseResult.Status.STOPPED, commitToPick);
            
            RebaseResult result = cherryPickCommit(commitToPick);
            if (result != null)
                return result;
            
        } 
        boolean isSquash = false;
        switch (step.getAction()) {
            case PICK :
                return null;
            case REWORD :
                String oldMessage = commitToPick.getFullMessage();
                String newMessage = interactiveHandler.modifyCommitMessage(oldMessage);
                newHead = new Git(repo).commit().setMessage(newMessage).setAmend(true).call();
                return null;
            case EDIT :
                rebaseState.createFile(AMEND, commitToPick.name());
                return stop(commitToPick, RebaseResult.Status.EDIT);
            case COMMENT :
                break;
            case SQUASH :
                isSquash = true;
            case FIXUP :
                resetSoftToParent();
                List<org.eclipse.jgit.lib.RebaseTodoLine> steps = repo.readRebaseTodo(rebaseState.getPath(GIT_REBASE_TODO), false);
                RebaseTodoLine nextStep = (steps.size()) > 0 ? steps.get(0) : null;
                File messageFixupFile = rebaseState.getFile(MESSAGE_FIXUP);
                File messageSquashFile = rebaseState.getFile(MESSAGE_SQUASH);
                if (isSquash && (messageFixupFile.exists()))
                    messageFixupFile.delete();
                
                newHead = doSquashFixup(isSquash, commitToPick, nextStep, messageFixupFile, messageSquashFile);
        }
        return null;
    }

    private RebaseResult cherryPickCommit(RevCommit commitToPick) throws IOException, ConcurrentRefUpdateException, GitAPIException, NoHeadException, NoMessageException, UnmergedPathsException, WrongRepositoryStateException {
        try {
            monitor.beginTask(java.text.MessageFormat.format(JGitText.get().applyingCommit, commitToPick.getShortMessage()), ProgressMonitor.UNKNOWN);
            if (preserveMerges)
                return cherryPickCommitPreservingMerges(commitToPick);
            else
                return cherryPickCommitFlattening(commitToPick);
            
        } finally {
            monitor.endTask();
        }
    }

    private RebaseResult cherryPickCommitFlattening(RevCommit commitToPick) throws IOException, ConcurrentRefUpdateException, GitAPIException, NoHeadException, NoMessageException, UnmergedPathsException, WrongRepositoryStateException {
        newHead = tryFastForward(commitToPick);
        lastStepWasForward = (newHead) != null;
        if (!(lastStepWasForward)) {
            String ourCommitName = getOurCommitName();
            CherryPickResult cherryPickResult = new Git(repo).cherryPick().include(commitToPick).setOurCommitName(ourCommitName).setReflogPrefix(REFLOG_PREFIX).setStrategy(strategy).call();
            switch (cherryPickResult.getStatus()) {
                case FAILED :
                    if ((operation) == (Operation.BEGIN))
                        return abort(RebaseResult.failed(cherryPickResult.getFailingPaths()));
                    else
                        return stop(commitToPick, RebaseResult.Status.STOPPED);
                    
                case CONFLICTING :
                    return stop(commitToPick, RebaseResult.Status.STOPPED);
                case OK :
                    newHead = cherryPickResult.getNewHead();
            }
        } 
        return null;
    }

    private RebaseResult cherryPickCommitPreservingMerges(RevCommit commitToPick) throws IOException, ConcurrentRefUpdateException, GitAPIException, NoHeadException, NoMessageException, UnmergedPathsException, WrongRepositoryStateException {
        writeCurrentCommit(commitToPick);
        List<org.eclipse.jgit.revwalk.RevCommit> newParents = getNewParents(commitToPick);
        boolean otherParentsUnchanged = true;
        for (int i = 1 ; i < (commitToPick.getParentCount()) ; i++)
            otherParentsUnchanged &= newParents.get(i).equals(commitToPick.getParent(i));
        newHead = otherParentsUnchanged ? tryFastForward(commitToPick) : null;
        lastStepWasForward = (newHead) != null;
        if (!(lastStepWasForward)) {
            ObjectId headId = getHead().getObjectId();
            if (!(AnyObjectId.equals(headId, newParents.get(0))))
                checkoutCommit(headId.getName(), newParents.get(0));
            
            if (otherParentsUnchanged) {
                boolean isMerge = (commitToPick.getParentCount()) > 1;
                String ourCommitName = getOurCommitName();
                CherryPickCommand pickCommand = new Git(repo).cherryPick().include(commitToPick).setOurCommitName(ourCommitName).setReflogPrefix(REFLOG_PREFIX).setStrategy(strategy);
                if (isMerge) {
                    pickCommand.setMainlineParentNumber(1);
                    pickCommand.setNoCommit(true);
                    writeMergeInfo(commitToPick, newParents);
                } 
                CherryPickResult cherryPickResult = pickCommand.call();
                switch (cherryPickResult.getStatus()) {
                    case FAILED :
                        if ((operation) == (Operation.BEGIN))
                            return abort(RebaseResult.failed(cherryPickResult.getFailingPaths()));
                        else
                            return stop(commitToPick, RebaseResult.Status.STOPPED);
                        
                    case CONFLICTING :
                        return stop(commitToPick, RebaseResult.Status.STOPPED);
                    case OK :
                        if (isMerge) {
                            CommitCommand commit = new Git(repo).commit();
                            commit.setAuthor(commitToPick.getAuthorIdent());
                            commit.setReflogComment((((REFLOG_PREFIX) + " ") + (commitToPick.getShortMessage())));
                            newHead = commit.call();
                        } else
                            newHead = cherryPickResult.getNewHead();
                        
                        break;
                }
            } else {
                MergeCommand merge = new Git(repo).merge().setFastForward(MergeCommand.FastForwardMode.NO_FF).setCommit(false);
                for (int i = 1 ; i < (commitToPick.getParentCount()) ; i++)
                    merge.include(newParents.get(i));
                MergeResult mergeResult = merge.call();
                if (mergeResult.getMergeStatus().isSuccessful()) {
                    CommitCommand commit = new Git(repo).commit();
                    commit.setAuthor(commitToPick.getAuthorIdent());
                    commit.setMessage(commitToPick.getFullMessage());
                    commit.setReflogComment((((REFLOG_PREFIX) + " ") + (commitToPick.getShortMessage())));
                    newHead = commit.call();
                } else {
                    if (((operation) == (Operation.BEGIN)) && ((mergeResult.getMergeStatus()) == (MergeResult.MergeStatus.FAILED)))
                        return abort(RebaseResult.failed(mergeResult.getFailingPaths()));
                    
                    return stop(commitToPick, RebaseResult.Status.STOPPED);
                }
            }
        } 
        return null;
    }

    private void writeMergeInfo(RevCommit commitToPick, List<org.eclipse.jgit.revwalk.RevCommit> newParents) throws IOException {
        repo.writeMergeHeads(newParents.subList(1, newParents.size()));
        repo.writeMergeCommitMsg(commitToPick.getFullMessage());
    }

    private List<org.eclipse.jgit.revwalk.RevCommit> getNewParents(RevCommit commitToPick) throws IOException {
        List<org.eclipse.jgit.revwalk.RevCommit> newParents = new ArrayList<org.eclipse.jgit.revwalk.RevCommit>();
        for (int p = 0 ; p < (commitToPick.getParentCount()) ; p++) {
            String parentHash = commitToPick.getParent(p).getName();
            if (!(new File(rebaseState.getRewrittenDir() , parentHash).exists()))
                newParents.add(commitToPick.getParent(p));
            else {
                String newParent = RebaseState.readFile(rebaseState.getRewrittenDir(), parentHash);
                if ((newParent.length()) == 0)
                    newParents.add(walk.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD)));
                else
                    newParents.add(walk.parseCommit(ObjectId.fromString(newParent)));
                
            }
        }
        return newParents;
    }

    private void writeCurrentCommit(RevCommit commit) throws IOException {
        RebaseState.appendToFile(rebaseState.getFile(CURRENT_COMMIT), commit.name());
    }

    private void writeRewrittenHashes() throws IOException, RevisionSyntaxException {
        File currentCommitFile = rebaseState.getFile(CURRENT_COMMIT);
        if (!(currentCommitFile.exists()))
            return ;
        
        String head = repo.resolve(org.eclipse.jgit.lib.Constants.HEAD).getName();
        String currentCommits = rebaseState.readFile(CURRENT_COMMIT);
        for (String current : currentCommits.split("\n"))
            RebaseState.createFile(rebaseState.getRewrittenDir(), current, head);
        org.eclipse.jgit.util.FileUtils.delete(currentCommitFile);
    }

    private RebaseResult finishRebase(RevCommit finalHead, boolean lastStepIsForward) throws IOException, GitAPIException {
        String headName = rebaseState.readFile(HEAD_NAME);
        updateHead(headName, finalHead, upstreamCommit);
        boolean stashConflicts = autoStashApply();
        org.eclipse.jgit.util.FileUtils.delete(rebaseState.getDir(), org.eclipse.jgit.util.FileUtils.RECURSIVE);
        if (stashConflicts)
            return RebaseResult.STASH_APPLY_CONFLICTS_RESULT;
        
        if (lastStepIsForward || (finalHead == null))
            return RebaseResult.FAST_FORWARD_RESULT;
        
        return RebaseResult.OK_RESULT;
    }

    private void checkSteps(List<org.eclipse.jgit.lib.RebaseTodoLine> steps) throws IOException, InvalidRebaseStepException {
        if (steps.isEmpty())
            return ;
        
        if ((RebaseTodoLine.Action.SQUASH.equals(steps.get(0).getAction())) || (RebaseTodoLine.Action.FIXUP.equals(steps.get(0).getAction()))) {
            if ((!(rebaseState.getFile(DONE).exists())) || ((rebaseState.readFile(DONE).trim().length()) == 0)) {
                throw new InvalidRebaseStepException(java.text.MessageFormat.format(JGitText.get().cannotSquashFixupWithoutPreviousCommit, steps.get(0).getAction().name()));
            } 
        } 
    }

    private RevCommit doSquashFixup(boolean isSquash, RevCommit commitToPick, RebaseTodoLine nextStep, File messageFixup, File messageSquash) throws IOException, GitAPIException {
        if (!(messageSquash.exists())) {
            ObjectId headId = repo.resolve(org.eclipse.jgit.lib.Constants.HEAD);
            RevCommit previousCommit = walk.parseCommit(headId);
            initializeSquashFixupFile(MESSAGE_SQUASH, previousCommit.getFullMessage());
            if (!isSquash)
                initializeSquashFixupFile(MESSAGE_FIXUP, previousCommit.getFullMessage());
            
        } 
        String currSquashMessage = rebaseState.readFile(MESSAGE_SQUASH);
        int count = (RebaseCommand.parseSquashFixupSequenceCount(currSquashMessage)) + 1;
        String content = RebaseCommand.composeSquashMessage(isSquash, commitToPick, currSquashMessage, count);
        rebaseState.createFile(MESSAGE_SQUASH, content);
        if (messageFixup.exists())
            rebaseState.createFile(MESSAGE_FIXUP, content);
        
        return squashIntoPrevious(!(messageFixup.exists()), nextStep);
    }

    private void resetSoftToParent() throws IOException, CheckoutConflictException, GitAPIException {
        Ref orig_head = repo.getRef(org.eclipse.jgit.lib.Constants.ORIG_HEAD);
        ObjectId orig_headId = orig_head.getObjectId();
        try {
            Git.wrap(repo).reset().setMode(org.eclipse.jgit.api.ResetCommand.ResetType.SOFT).setRef("HEAD~1").call();
        } finally {
            repo.writeOrigHead(orig_headId);
        }
    }

    private RevCommit squashIntoPrevious(boolean sequenceContainsSquash, RebaseTodoLine nextStep) throws IOException, GitAPIException {
        RevCommit retNewHead;
        String commitMessage = rebaseState.readFile(MESSAGE_SQUASH);
        if ((nextStep == null) || (((nextStep.getAction()) != (RebaseTodoLine.Action.FIXUP)) && ((nextStep.getAction()) != (RebaseTodoLine.Action.SQUASH)))) {
            if (sequenceContainsSquash) {
                commitMessage = interactiveHandler.modifyCommitMessage(commitMessage);
            } 
            retNewHead = new Git(repo).commit().setMessage(RebaseCommand.stripCommentLines(commitMessage)).setAmend(true).call();
            rebaseState.getFile(MESSAGE_SQUASH).delete();
            rebaseState.getFile(MESSAGE_FIXUP).delete();
        } else {
            retNewHead = new Git(repo).commit().setMessage(commitMessage).setAmend(true).call();
        }
        return retNewHead;
    }

    private static String stripCommentLines(String commitMessage) {
        StringBuilder result = new StringBuilder();
        for (String line : commitMessage.split("\n")) {
            if (!(line.trim().startsWith("#")))
                result.append(line).append("\n");
            
        }
        if (!(commitMessage.endsWith("\n")))
            result.deleteCharAt(((result.length()) - 1));
        
        return result.toString();
    }

    @SuppressWarnings(value = "nls")
    private static String composeSquashMessage(boolean isSquash, RevCommit commitToPick, String currSquashMessage, int count) {
        StringBuilder sb = new StringBuilder();
        String ordinal = RebaseCommand.getOrdinal(count);
        sb.setLength(0);
        sb.append("# This is a combination of ").append(count).append(" commits.\n");
        sb.append(currSquashMessage.substring(((currSquashMessage.indexOf("\n")) + 1)));
        sb.append("\n");
        if (isSquash) {
            sb.append("# This is the ").append(count).append(ordinal).append(" commit message:\n");
            sb.append(commitToPick.getFullMessage());
        } else {
            sb.append("# The ").append(count).append(ordinal).append(" commit message will be skipped:\n# ");
            sb.append(commitToPick.getFullMessage().replaceAll("([\n\r])", "$1# "));
        }
        return sb.toString();
    }

    private static String getOrdinal(int count) {
        switch (count % 10) {
            case 1 :
                return "st";
            case 2 :
                return "nd";
            case 3 :
                return "rd";
            default :
                return "th";
        }
    }

    /** 
     * * Parse the count from squashed commit messages
     * *
     * * @param currSquashMessage
     * *            the squashed commit message to be parsed
     * * @return the count of squashed messages in the given string
     */
static int parseSquashFixupSequenceCount(String currSquashMessage) {
        String regex = "This is a combination of (.*) commits";
        String firstLine = currSquashMessage.substring(0, currSquashMessage.indexOf("\n"));
        Pattern pattern = Pattern.compile(regex);
        Matcher matcher = pattern.matcher(firstLine);
        if (!(matcher.find()))
            throw new IllegalArgumentException();
        
        return java.lang.Integer.parseInt(matcher.group(1));
    }

    private void initializeSquashFixupFile(String messageFile, String fullMessage) throws IOException {
        rebaseState.createFile(messageFile, ("# This is a combination of 1 commits.\n# The first commit\'s message is:\n" + fullMessage));
    }

    private String getOurCommitName() {
        String ourCommitName = "Upstream, based on " + (Repository.shortenRefName(upstreamCommitName));
        return ourCommitName;
    }

    private void updateHead(String headName, RevCommit aNewHead, RevCommit onto) throws IOException {
        if (headName.startsWith(org.eclipse.jgit.lib.Constants.R_REFS)) {
            RefUpdate rup = repo.updateRef(headName);
            rup.setNewObjectId(aNewHead);
            rup.setRefLogMessage(((("rebase finished: " + headName) + " onto ") + (onto.getName())), false);
            RefUpdate.Result res = rup.forceUpdate();
            switch (res) {
                case FAST_FORWARD :
                case FORCED :
                case NO_CHANGE :
                    break;
                default :
                    throw new JGitInternalException("Updating HEAD failed");
            }
            rup = repo.updateRef(org.eclipse.jgit.lib.Constants.HEAD);
            rup.setRefLogMessage(("rebase finished: returning to " + headName), false);
            res = rup.link(headName);
            switch (res) {
                case FAST_FORWARD :
                case FORCED :
                case NO_CHANGE :
                    break;
                default :
                    throw new JGitInternalException("Updating HEAD failed");
            }
        } 
    }

    private RevCommit checkoutCurrentHead() throws IOException, NoHeadException {
        ObjectId headTree = repo.resolve(((org.eclipse.jgit.lib.Constants.HEAD) + "^{tree}"));
        if (headTree == null)
            throw new NoHeadException(JGitText.get().cannotRebaseWithoutCurrentHead);
        
        DirCache dc = repo.lockDirCache();
        try {
            DirCacheCheckout dco = new DirCacheCheckout(repo , dc , headTree);
            dco.setFailOnConflict(false);
            boolean needsDeleteFiles = dco.checkout();
            if (needsDeleteFiles) {
                List<java.lang.String> fileList = dco.getToBeDeleted();
                for (String filePath : fileList) {
                    File fileToDelete = new File(repo.getWorkTree() , filePath);
                    if (repo.getFS().exists(fileToDelete))
                        org.eclipse.jgit.util.FileUtils.delete(fileToDelete, ((org.eclipse.jgit.util.FileUtils.RECURSIVE) | (org.eclipse.jgit.util.FileUtils.RETRY)));
                    
                }
            } 
        } finally {
            dc.unlock();
        }
        RevWalk rw = new RevWalk(repo);
        RevCommit commit = rw.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD));
        rw.release();
        return commit;
    }

    /** 
     * * @return the commit if we had to do a commit, otherwise null
     * * @throws GitAPIException
     * * @throws IOException
     */
private RevCommit continueRebase() throws IOException, GitAPIException {
        DirCache dc = repo.readDirCache();
        boolean hasUnmergedPaths = dc.hasUnmergedPaths();
        if (hasUnmergedPaths)
            throw new UnmergedPathsException();
        
        TreeWalk treeWalk = new TreeWalk(repo);
        treeWalk.reset();
        treeWalk.setRecursive(true);
        treeWalk.addTree(new org.eclipse.jgit.dircache.DirCacheIterator(dc));
        ObjectId id = repo.resolve(((org.eclipse.jgit.lib.Constants.HEAD) + "^{tree}"));
        if (id == null)
            throw new NoHeadException(JGitText.get().cannotRebaseWithoutCurrentHead);
        
        treeWalk.addTree(id);
        treeWalk.setFilter(org.eclipse.jgit.treewalk.filter.TreeFilter.ANY_DIFF);
        boolean needsCommit = treeWalk.next();
        treeWalk.release();
        if (needsCommit) {
            CommitCommand commit = new Git(repo).commit();
            commit.setMessage(rebaseState.readFile(MESSAGE));
            commit.setAuthor(parseAuthor());
            return commit.call();
        } 
        return null;
    }

    private PersonIdent parseAuthor() throws IOException {
        File authorScriptFile = rebaseState.getFile(AUTHOR_SCRIPT);
        byte[] raw;
        try {
            raw = org.eclipse.jgit.util.IO.readFully(authorScriptFile);
        } catch (FileNotFoundException notFound) {
            return null;
        }
        return parseAuthor(raw);
    }

    private RebaseResult stop(RevCommit commitToPick, RebaseResult.Status status) throws IOException {
        PersonIdent author = commitToPick.getAuthorIdent();
        String authorScript = toAuthorScript(author);
        rebaseState.createFile(AUTHOR_SCRIPT, authorScript);
        rebaseState.createFile(MESSAGE, commitToPick.getFullMessage());
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        DiffFormatter df = new DiffFormatter(bos);
        df.setRepository(repo);
        df.format(commitToPick.getParent(0), commitToPick);
        rebaseState.createFile(PATCH, new String(bos.toByteArray() , org.eclipse.jgit.lib.Constants.CHARACTER_ENCODING));
        if (!(Repository.isValidRefName(("x/" + STOPPED_SHA))))
            throw new RevisionSyntaxException(STOPPED_SHA);
        
        repo.writeCherryPickHead(null);
        return RebaseResult.result(status, commitToPick);
    }

    String toAuthorScript(PersonIdent author) {
        StringBuilder sb = new StringBuilder(100);
        sb.append(GIT_AUTHOR_NAME);
        sb.append("=\'");
        sb.append(author.getName());
        sb.append("\'\n");
        sb.append(GIT_AUTHOR_EMAIL);
        sb.append("=\'");
        sb.append(author.getEmailAddress());
        sb.append("\'\n");
        sb.append(GIT_AUTHOR_DATE);
        sb.append("=\'");
        sb.append("@");
        String externalString = author.toExternalString();
        sb.append(externalString.substring(((externalString.lastIndexOf('>')) + 2)));
        sb.append("\'\n");
        return sb.toString();
    }

    /** 
     * * Removes the number of lines given in the parameter from the
     * * <code>git-rebase-todo</code> file but preserves comments and other lines
     * * that can not be parsed as steps
     * *
     * * @param numSteps
     * * @throws IOException
     */
private void popSteps(int numSteps) throws IOException {
        if (numSteps == 0)
            return ;
        
        List<org.eclipse.jgit.lib.RebaseTodoLine> todoLines = new LinkedList<org.eclipse.jgit.lib.RebaseTodoLine>();
        List<org.eclipse.jgit.lib.RebaseTodoLine> poppedLines = new LinkedList<org.eclipse.jgit.lib.RebaseTodoLine>();
        for (RebaseTodoLine line : repo.readRebaseTodo(rebaseState.getPath(GIT_REBASE_TODO), true)) {
            if (((poppedLines.size()) >= numSteps) || (RebaseTodoLine.Action.COMMENT.equals(line.getAction())))
                todoLines.add(line);
            else
                poppedLines.add(line);
            
        }
        repo.writeRebaseTodoFile(rebaseState.getPath(GIT_REBASE_TODO), todoLines, false);
        if ((poppedLines.size()) > 0) {
            repo.writeRebaseTodoFile(rebaseState.getPath(DONE), poppedLines, true);
        } 
    }

    private RebaseResult initFilesAndRewind() throws IOException, GitAPIException {
        Ref head = getHead();
        String headName = RebaseCommand.getHeadName(head);
        ObjectId headId = head.getObjectId();
        if (headId == null)
            throw new RefNotFoundException(java.text.MessageFormat.format(JGitText.get().refNotResolved, org.eclipse.jgit.lib.Constants.HEAD));
        
        RevCommit headCommit = walk.lookupCommit(headId);
        RevCommit upstream = walk.lookupCommit(upstreamCommit.getId());
        if ((!(isInteractive())) && (walk.isMergedInto(upstream, headCommit)))
            return RebaseResult.UP_TO_DATE_RESULT;
        else if ((!(isInteractive())) && (walk.isMergedInto(headCommit, upstream))) {
            monitor.beginTask(java.text.MessageFormat.format(JGitText.get().resettingHead, upstreamCommit.getShortMessage()), ProgressMonitor.UNKNOWN);
            checkoutCommit(headName, upstreamCommit);
            monitor.endTask();
            updateHead(headName, upstreamCommit, upstream);
            return RebaseResult.FAST_FORWARD_RESULT;
        } 
        monitor.beginTask(JGitText.get().obtainingCommitsForCherryPick, ProgressMonitor.UNKNOWN);
        org.eclipse.jgit.util.FileUtils.mkdir(rebaseState.getDir(), true);
        repo.writeOrigHead(headId);
        rebaseState.createFile(REBASE_HEAD, headId.name());
        rebaseState.createFile(HEAD_NAME, headName);
        rebaseState.createFile(ONTO, upstreamCommit.name());
        rebaseState.createFile(ONTO_NAME, upstreamCommitName);
        rebaseState.createFile(INTERACTIVE, "");
        rebaseState.createFile(QUIET, "");
        ArrayList<org.eclipse.jgit.lib.RebaseTodoLine> toDoSteps = new ArrayList<org.eclipse.jgit.lib.RebaseTodoLine>();
        toDoSteps.add(new RebaseTodoLine(((("# Created by EGit: rebasing " + (headId.name())) + " onto ") + (upstreamCommit.name()))));
        List<org.eclipse.jgit.revwalk.RevCommit> cherryPickList = calculatePickList(headCommit);
        ObjectReader reader = walk.getObjectReader();
        for (RevCommit commit : cherryPickList)
            toDoSteps.add(new RebaseTodoLine(RebaseTodoLine.Action.PICK , reader.abbreviate(commit) , commit.getShortMessage()));
        repo.writeRebaseTodoFile(rebaseState.getPath(GIT_REBASE_TODO), toDoSteps, false);
        monitor.endTask();
        monitor.beginTask(java.text.MessageFormat.format(JGitText.get().rewinding, upstreamCommit.getShortMessage()), ProgressMonitor.UNKNOWN);
        boolean checkoutOk = false;
        try {
            checkoutOk = checkoutCommit(headName, upstreamCommit);
        } finally {
            if (!checkoutOk)
                org.eclipse.jgit.util.FileUtils.delete(rebaseState.getDir(), org.eclipse.jgit.util.FileUtils.RECURSIVE);
            
        }
        monitor.endTask();
        return null;
    }

    private List<org.eclipse.jgit.revwalk.RevCommit> calculatePickList(RevCommit headCommit) throws IOException, GitAPIException, NoHeadException {
        LogCommand cmd = new Git(repo).log().addRange(upstreamCommit, headCommit);
        Iterable<org.eclipse.jgit.revwalk.RevCommit> commitsToUse = cmd.call();
        List<org.eclipse.jgit.revwalk.RevCommit> cherryPickList = new ArrayList<org.eclipse.jgit.revwalk.RevCommit>();
        for (RevCommit commit : commitsToUse) {
            if ((preserveMerges) || ((commit.getParentCount()) == 1))
                cherryPickList.add(commit);
            
        }
        java.util.Collections.reverse(cherryPickList);
        if (preserveMerges) {
            File rewrittenDir = rebaseState.getRewrittenDir();
            org.eclipse.jgit.util.FileUtils.mkdir(rewrittenDir, false);
            walk.reset();
            walk.setRevFilter(org.eclipse.jgit.revwalk.filter.RevFilter.MERGE_BASE);
            walk.markStart(upstreamCommit);
            walk.markStart(headCommit);
            RevCommit base;
            while ((base = walk.next()) != null)
                RebaseState.createFile(rewrittenDir, base.getName(), upstreamCommit.getName());
            Iterator<org.eclipse.jgit.revwalk.RevCommit> iterator = cherryPickList.iterator();
            pickLoop : while (iterator.hasNext()) {
                RevCommit commit = iterator.next();
                for (int i = 0 ; i < (commit.getParentCount()) ; i++) {
                    boolean parentRewritten = new File(rewrittenDir , commit.getParent(i).getName()).exists();
                    if (parentRewritten) {
                        new File(rewrittenDir , commit.getName()).createNewFile();
                        continue pickLoop;
                    } 
                }
                iterator.remove();
            }
        } 
        return cherryPickList;
    }

    private static String getHeadName(Ref head) {
        String headName;
        if (head.isSymbolic())
            headName = head.getTarget().getName();
        else
            headName = head.getObjectId().getName();
        
        return headName;
    }

    private Ref getHead() throws IOException, RefNotFoundException {
        Ref head = repo.getRef(org.eclipse.jgit.lib.Constants.HEAD);
        if ((head == null) || ((head.getObjectId()) == null))
            throw new RefNotFoundException(java.text.MessageFormat.format(JGitText.get().refNotResolved, org.eclipse.jgit.lib.Constants.HEAD));
        
        return head;
    }

    private boolean isInteractive() {
        return (interactiveHandler) != null;
    }

    /** 
     * * checks if we can fast-forward and returns the new head if it is possible
     * *
     * * @param newCommit
     * * @return the new head, or null
     * * @throws IOException
     * * @throws GitAPIException
     */
public RevCommit tryFastForward(RevCommit newCommit) throws IOException, GitAPIException {
        Ref head = getHead();
        ObjectId headId = head.getObjectId();
        if (headId == null)
            throw new RefNotFoundException(java.text.MessageFormat.format(JGitText.get().refNotResolved, org.eclipse.jgit.lib.Constants.HEAD));
        
        RevCommit headCommit = walk.lookupCommit(headId);
        if (walk.isMergedInto(newCommit, headCommit))
            return newCommit;
        
        String headName = RebaseCommand.getHeadName(head);
        return tryFastForward(headName, headCommit, newCommit);
    }

    private RevCommit tryFastForward(String headName, RevCommit oldCommit, RevCommit newCommit) throws IOException, GitAPIException {
        boolean tryRebase = false;
        for (RevCommit parentCommit : newCommit.getParents())
            if (parentCommit.equals(oldCommit))
                tryRebase = true;
            
        if (!tryRebase)
            return null;
        
        CheckoutCommand co = new CheckoutCommand(repo);
        try {
            co.setName(newCommit.name()).call();
            if (headName.startsWith(org.eclipse.jgit.lib.Constants.R_HEADS)) {
                RefUpdate rup = repo.updateRef(headName);
                rup.setExpectedOldObjectId(oldCommit);
                rup.setNewObjectId(newCommit);
                rup.setRefLogMessage(((("Fast-foward from " + (oldCommit.name())) + " to ") + (newCommit.name())), false);
                RefUpdate.Result res = rup.update(walk);
                switch (res) {
                    case FAST_FORWARD :
                    case NO_CHANGE :
                    case FORCED :
                        break;
                    default :
                        throw new IOException("Could not fast-forward");
                }
            } 
            return newCommit;
        } catch (RefAlreadyExistsException e) {
            throw new JGitInternalException(e.getMessage() , e);
        } catch (RefNotFoundException e) {
            throw new JGitInternalException(e.getMessage() , e);
        } catch (InvalidRefNameException e) {
            throw new JGitInternalException(e.getMessage() , e);
        } catch (CheckoutConflictException e) {
            throw new JGitInternalException(e.getMessage() , e);
        }
    }

    private void checkParameters() throws WrongRepositoryStateException {
        if ((this.operation) == (Operation.PROCESS_STEPS)) {
            if (rebaseState.getFile(DONE).exists())
                throw new WrongRepositoryStateException(java.text.MessageFormat.format(JGitText.get().wrongRepositoryState, repo.getRepositoryState().name()));
            
        } 
        if ((this.operation) != (Operation.BEGIN)) {
            switch (repo.getRepositoryState()) {
                case REBASING_INTERACTIVE :
                case REBASING :
                case REBASING_REBASING :
                case REBASING_MERGE :
                    break;
                default :
                    throw new WrongRepositoryStateException(java.text.MessageFormat.format(JGitText.get().wrongRepositoryState, repo.getRepositoryState().name()));
            }
        } else
            switch (repo.getRepositoryState()) {
                case SAFE :
                    if ((this.upstreamCommit) == null)
                        throw new JGitInternalException(java.text.MessageFormat.format(JGitText.get().missingRequiredParameter, "upstream"));
                    
                    return ;
                default :
                    throw new WrongRepositoryStateException(java.text.MessageFormat.format(JGitText.get().wrongRepositoryState, repo.getRepositoryState().name()));
            }
        
    }

    private RebaseResult abort(RebaseResult result) throws IOException, GitAPIException {
        try {
            ObjectId origHead = repo.readOrigHead();
            String commitId = origHead != null ? origHead.name() : null;
            monitor.beginTask(java.text.MessageFormat.format(JGitText.get().abortingRebase, commitId), ProgressMonitor.UNKNOWN);
            DirCacheCheckout dco;
            if (commitId == null)
                throw new JGitInternalException(JGitText.get().abortingRebaseFailedNoOrigHead);
            
            ObjectId id = repo.resolve(commitId);
            RevCommit commit = walk.parseCommit(id);
            if (result.getStatus().equals(RebaseResult.Status.FAILED)) {
                RevCommit head = walk.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD));
                dco = new DirCacheCheckout(repo , head.getTree() , repo.lockDirCache() , commit.getTree());
            } else {
                dco = new DirCacheCheckout(repo , repo.lockDirCache() , commit.getTree());
            }
            dco.setFailOnConflict(false);
            dco.checkout();
            walk.release();
        } finally {
            monitor.endTask();
        }
        try {
            String headName = rebaseState.readFile(HEAD_NAME);
            monitor.beginTask(java.text.MessageFormat.format(JGitText.get().resettingHead, headName), ProgressMonitor.UNKNOWN);
            RefUpdate.Result res = null;
            RefUpdate refUpdate = repo.updateRef(org.eclipse.jgit.lib.Constants.HEAD, false);
            refUpdate.setRefLogMessage("rebase: aborting", false);
            if (headName.startsWith(org.eclipse.jgit.lib.Constants.R_REFS)) {
                res = refUpdate.link(headName);
            } else {
                refUpdate.setNewObjectId(repo.readOrigHead());
                res = refUpdate.forceUpdate();
            }
            switch (res) {
                case FAST_FORWARD :
                case FORCED :
                case NO_CHANGE :
                    break;
                default :
                    throw new JGitInternalException(JGitText.get().abortingRebaseFailed);
            }
            boolean stashConflicts = autoStashApply();
            org.eclipse.jgit.util.FileUtils.delete(rebaseState.getDir(), org.eclipse.jgit.util.FileUtils.RECURSIVE);
            repo.writeCherryPickHead(null);
            repo.writeMergeHeads(null);
            if (stashConflicts)
                return RebaseResult.STASH_APPLY_CONFLICTS_RESULT;
            
            return result;
        } finally {
            monitor.endTask();
        }
    }

    private boolean checkoutCommit(String headName, RevCommit commit) throws IOException, CheckoutConflictException {
        try {
            RevCommit head = walk.parseCommit(repo.resolve(org.eclipse.jgit.lib.Constants.HEAD));
            DirCacheCheckout dco = new DirCacheCheckout(repo , head.getTree() , repo.lockDirCache() , commit.getTree());
            dco.setFailOnConflict(true);
            try {
                dco.checkout();
            } catch (org.eclipse.jgit.errors.CheckoutConflictException cce) {
                throw new CheckoutConflictException(dco.getConflicts() , cce);
            }
            RefUpdate refUpdate = repo.updateRef(org.eclipse.jgit.lib.Constants.HEAD, true);
            refUpdate.setExpectedOldObjectId(head);
            refUpdate.setNewObjectId(commit);
            refUpdate.setRefLogMessage(((("checkout: moving from " + (Repository.shortenRefName(headName))) + " to ") + (commit.getName())), false);
            RefUpdate.Result res = refUpdate.forceUpdate();
            switch (res) {
                case FAST_FORWARD :
                case NO_CHANGE :
                case FORCED :
                    break;
                default :
                    throw new IOException("Could not rewind to upstream commit");
            }
        } finally {
            walk.release();
            monitor.endTask();
        }
        return true;
    }

    /** 
     * * @param upstream
     * *            the upstream commit
     * * @return {@code this}
     */
public RebaseCommand setUpstream(RevCommit upstream) {
        this.upstreamCommit = upstream;
        this.upstreamCommitName = upstream.name();
        return this;
    }

    /** 
     * * @param upstream
     * *            id of the upstream commit
     * * @return {@code this}
     */
public RebaseCommand setUpstream(AnyObjectId upstream) {
        try {
            this.upstreamCommit = walk.parseCommit(upstream);
            this.upstreamCommitName = upstream.name();
        } catch (IOException e) {
            throw new JGitInternalException(java.text.MessageFormat.format(JGitText.get().couldNotReadObjectWhileParsingCommit, upstream.name()) , e);
        }
        return this;
    }

    /** 
     * * @param upstream
     * *            the upstream branch
     * * @return {@code this}
     * * @throws RefNotFoundException
     */
public RebaseCommand setUpstream(String upstream) throws RefNotFoundException {
        try {
            ObjectId upstreamId = repo.resolve(upstream);
            if (upstreamId == null)
                throw new RefNotFoundException(java.text.MessageFormat.format(JGitText.get().refNotResolved, upstream));
            
            upstreamCommit = walk.parseCommit(repo.resolve(upstream));
            upstreamCommitName = upstream;
            return this;
        } catch (IOException ioe) {
            throw new JGitInternalException(ioe.getMessage() , ioe);
        }
    }

    /** 
     * * Optionally override the name of the upstream. If this is used, it has to
     * * come after any {@link #setUpstream} call.
     * *
     * * @param upstreamName
     * *            the name which will be used to refer to upstream in conflicts
     * * @return {@code this}
     */
public RebaseCommand setUpstreamName(String upstreamName) {
        if ((upstreamCommit) == null) {
            throw new IllegalStateException("setUpstreamName must be called after setUpstream.");
        } 
        this.upstreamCommitName = upstreamName;
        return this;
    }

    /** 
     * * @param operation
     * *            the operation to perform
     * * @return {@code this}
     */
public RebaseCommand setOperation(Operation operation) {
        this.operation = operation;
        return this;
    }

    /** 
     * * @param monitor
     * *            a progress monitor
     * * @return this instance
     */
public RebaseCommand setProgressMonitor(ProgressMonitor monitor) {
        this.monitor = monitor;
        return this;
    }

    /** 
     * * Enables interactive rebase
     * * <p>
     * * Does not stop after initialization of interactive rebase. This is
     * * equivalent to
     * * {@link RebaseCommand#runInteractively(InteractiveHandler, boolean)
     * * runInteractively(handler, false)};
     * * </p>
     * *
     * * @param handler
     * * @return this
     */
public RebaseCommand runInteractively(InteractiveHandler handler) {
        return runInteractively(handler, false);
    }

    /** 
     * * Enables interactive rebase
     * * <p>
     * * If stopAfterRebaseInteractiveInitialization is {@code true} the rebase
     * * stops after initialization of interactive rebase returning
     * * {@link RebaseResult#INTERACTIVE_PREPARED_RESULT}
     * * </p>
     * *
     * * @param handler
     * * @param stopAfterRebaseInteractiveInitialization
     * *            if {@code true} the rebase stops after initialization
     * * @return this instance
     * * @since 3.2
     */
public RebaseCommand runInteractively(InteractiveHandler handler, final boolean stopAfterRebaseInteractiveInitialization) {
        this.stopAfterInitialization = stopAfterRebaseInteractiveInitialization;
        this.interactiveHandler = handler;
        return this;
    }

    /** 
     * * @param strategy
     * *            The merge strategy to use during this rebase operation.
     * * @return {@code this}
     * * @since 3.4
     */
public RebaseCommand setStrategy(MergeStrategy strategy) {
        this.strategy = strategy;
        return this;
    }

    /** 
     * * @param preserve
     * *            True to re-create merges during rebase. Defaults to false, a
     * *            flattening rebase.
     * * @return {@code this}
     * * @since 3.5
     */
public RebaseCommand setPreserveMerges(boolean preserve) {
        this.preserveMerges = preserve;
        return this;
    }

    /** 
     * * Allows configure rebase interactive process and modify commit message
     */
public interface InteractiveHandler {
        /** 
         * * Given list of {@code steps} should be modified according to user
         * * rebase configuration
         * * @param steps
         * *            initial configuration of rebase interactive
         */
void prepareSteps(List<org.eclipse.jgit.lib.RebaseTodoLine> steps);

        /** 
         * * Used for editing commit message on REWORD
         * *
         * * @param commit
         * * @return new commit message
         */
String modifyCommitMessage(String commit);
    }

    PersonIdent parseAuthor(byte[] raw) {
        if ((raw.length) == 0)
            return null;
        
        Map<java.lang.String, java.lang.String> keyValueMap = new HashMap<java.lang.String, java.lang.String>();
        for (int p = 0 ; p < (raw.length) ; ) {
            int end = org.eclipse.jgit.util.RawParseUtils.nextLF(raw, p);
            if (end == p)
                break;
            
            int equalsIndex = org.eclipse.jgit.util.RawParseUtils.next(raw, p, '=');
            if (equalsIndex == end)
                break;
            
            String key = org.eclipse.jgit.util.RawParseUtils.decode(raw, p, (equalsIndex - 1));
            String value = org.eclipse.jgit.util.RawParseUtils.decode(raw, (equalsIndex + 1), (end - 2));
            p = end;
            keyValueMap.put(key, value);
        }
        String name = keyValueMap.get(GIT_AUTHOR_NAME);
        String email = keyValueMap.get(GIT_AUTHOR_EMAIL);
        String time = keyValueMap.get(GIT_AUTHOR_DATE);
        int timeStart = 0;
        if (time.startsWith("@"))
            timeStart = 1;
        else
            timeStart = 0;
        
        long when = (java.lang.Long.parseLong(time.substring(timeStart, time.indexOf(' ')))) * 1000;
        String tzOffsetString = time.substring(((time.indexOf(' ')) + 1));
        int multiplier = -1;
        if ((tzOffsetString.charAt(0)) == '+')
            multiplier = 1;
        
        int hours = java.lang.Integer.parseInt(tzOffsetString.substring(1, 3));
        int minutes = java.lang.Integer.parseInt(tzOffsetString.substring(3, 5));
        int tz = ((hours * 60) + minutes) * multiplier;
        if ((name != null) && (email != null))
            return new PersonIdent(name , email , when , tz);
        
        return null;
    }

    private static class RebaseState {
        private final File repoDirectory;

        private File dir;

        public RebaseState(File repoDirectory) {
            this.repoDirectory = repoDirectory;
        }

        public File getDir() {
            if ((dir) == null) {
                File rebaseApply = new File(repoDirectory , RebaseCommand.REBASE_APPLY);
                if (rebaseApply.exists()) {
                    dir = rebaseApply;
                } else {
                    File rebaseMerge = new File(repoDirectory , RebaseCommand.REBASE_MERGE);
                    dir = rebaseMerge;
                }
            } 
            return dir;
        }

        /** 
         * * @return Directory with rewritten commit hashes, usually exists if
         * *         {@link RebaseCommand#preserveMerges} is true
         */
public File getRewrittenDir() {
            return new File(getDir() , RebaseCommand.REWRITTEN);
        }

        public String readFile(String name) throws IOException {
            return RebaseState.readFile(getDir(), name);
        }

        public void createFile(String name, String content) throws IOException {
            RebaseState.createFile(getDir(), name, content);
        }

        public File getFile(String name) {
            return new File(getDir() , name);
        }

        public String getPath(String name) {
            return ((getDir().getName()) + "/") + name;
        }

        private static String readFile(File directory, String fileName) throws IOException {
            byte[] content = org.eclipse.jgit.util.IO.readFully(new File(directory , fileName));
            int end = org.eclipse.jgit.util.RawParseUtils.prevLF(content, content.length);
            return org.eclipse.jgit.util.RawParseUtils.decode(content, 0, (end + 1));
        }

        private static void createFile(File parentDir, String name, String content) throws IOException {
            File file = new File(parentDir , name);
            FileOutputStream fos = new FileOutputStream(file);
            try {
                fos.write(content.getBytes(org.eclipse.jgit.lib.Constants.CHARACTER_ENCODING));
                fos.write('\n');
            } finally {
                fos.close();
            }
        }

        private static void appendToFile(File file, String content) throws IOException {
            FileOutputStream fos = new FileOutputStream(file , true);
            try {
                fos.write(content.getBytes(org.eclipse.jgit.lib.Constants.CHARACTER_ENCODING));
                fos.write('\n');
            } finally {
                fos.close();
            }
        }
    }
}

