package org.eclipse.jgit.lib;

import org.eclipse.jgit.internal.JGitText;

/** 
 * Important state of the repository that affects what can and cannot bed
 * done. This is things like unhandled conflicted merges and unfinished rebase.
 * 
 * The granularity and set of states are somewhat arbitrary. The methods
 * on the state are the only supported means of deciding what to do.
 */
public enum RepositoryState {
BARE {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return false;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return "Bare";
        }
    }, SAFE {
        @Override
        public boolean canCheckout() {
            return true;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_normal;
        }
    }, MERGING {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return false;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_conflicts;
        }
    }, MERGING_RESOLVED {
        @Override
        public boolean canCheckout() {
            return true;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_merged;
        }
    }, CHERRY_PICKING {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return false;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_conflicts;
        }
    }, CHERRY_PICKING_RESOLVED {
        @Override
        public boolean canCheckout() {
            return true;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_merged;
        }
    }, REVERTING {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return false;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_conflicts;
        }
    }, REVERTING_RESOLVED {
        @Override
        public boolean canCheckout() {
            return true;
        }

        @Override
        public boolean canResetHead() {
            return true;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_merged;
        }
    }, REBASING {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return true;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_rebaseOrApplyMailbox;
        }
    }, REBASING_REBASING {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return true;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_rebase;
        }
    }, APPLY {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_applyMailbox;
        }
    }, REBASING_MERGE {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return true;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_rebaseWithMerge;
        }
    }, REBASING_INTERACTIVE {
        @Override
        public boolean canCheckout() {
            return false;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return true;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_rebaseInteractive;
        }
    }, BISECTING {
        @Override
        public boolean canCheckout() {
            return true;
        }

        @Override
        public boolean canResetHead() {
            return false;
        }

        @Override
        public boolean canCommit() {
            return true;
        }

        @Override
        public boolean canAmend() {
            return false;
        }

        @Override
        public boolean isRebasing() {
            return false;
        }

        @Override
        public String getDescription() {
            return JGitText.get().repositoryState_bisecting;
        }
    };
    /** 
     * * @return true if changing HEAD is sane.
     */
public abstract boolean canCheckout();

    /** 
     * * @return true if we can commit
     */
public abstract boolean canCommit();

    /** 
     * * @return true if reset to another HEAD is considered SAFE
     */
public abstract boolean canResetHead();

    /** 
     * * @return true if amending is considered SAFE
     */
public abstract boolean canAmend();

    /** 
     * * @return true if the repository is currently in a rebase
     * * @since 3.0
     */
public abstract boolean isRebasing();

    /** 
     * * @return a human readable description of the state.
     */
public abstract String getDescription();
}

