package org.eclipse.jgit.transport;

import org.eclipse.jgit.lib.AnyObjectId;
import java.util.Collection;
import org.eclipse.jgit.lib.Config;
import org.eclipse.jgit.lib.Constants;
import java.util.Date;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import org.eclipse.jgit.lib.MutableObjectId;
import org.eclipse.jgit.lib.NullProgressMonitor;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectInserter;
import java.io.OutputStream;
import org.eclipse.jgit.internal.storage.file.PackLock;
import org.eclipse.jgit.errors.PackProtocolException;
import org.eclipse.jgit.lib.ProgressMonitor;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevCommitList;
import org.eclipse.jgit.revwalk.RevFlag;
import org.eclipse.jgit.revwalk.RevObject;
import org.eclipse.jgit.revwalk.RevWalk;
import java.util.Set;
import org.eclipse.jgit.util.TemporaryBuffer;
import org.eclipse.jgit.errors.TransportException;

/** 
 * Fetch implementation using the native Git pack transfer service.
 * <p>
 * This is the canonical implementation for transferring objects from the remote
 * repository to the local repository by talking to the 'git-upload-pack'
 * service. Objects are packed on the remote side into a pack file and then sent
 * down the pipe to us.
 * <p>
 * This connection requires only a bi-directional pipe or socket, and thus is
 * easily wrapped up into a local process pipe, anonymous TCP socket, or a
 * command executed through an SSH tunnel.
 * <p>
 * If {@link BasePackConnection#statelessRPC} is {@code true}, this connection
 * can be tunneled over a request-response style RPC system like HTTP.  The RPC
 * call boundary is determined by this class switching from writing to the
 * OutputStream to reading from the InputStream.
 * <p>
 * Concrete implementations should just call
 * {@link #init(java.io.InputStream, java.io.OutputStream)} and
 * {@link #readAdvertisedRefs()} methods in constructor or before any use. They
 * should also handle resources releasing in {@link #close()} method if needed.
 */
public abstract class BasePackFetchConnection extends BasePackConnection implements FetchConnection {
    /** 
     * * Maximum number of 'have' lines to send before giving up.
     * * <p>
     * * During {@link #negotiate(ProgressMonitor)} we send at most this many
     * * commits to the remote peer as 'have' lines without an ACK response before
     * * we give up.
     */
private static final int MAX_HAVES = 256;

    /** 
     * * Amount of data the client sends before starting to read.
     * * <p>
     * * Any output stream given to the client must be able to buffer this many
     * * bytes before the client will stop writing and start reading from the
     * * input stream. If the output stream blocks before this many bytes are in
     * * the send queue, the system will deadlock.
     */
protected static final int MIN_CLIENT_BUFFER = ((2 * 32) * 46) + 8;

    /** 
     * * Include tags if we are also including the referenced objects.
     * * @since 2.0
     */
public static final String OPTION_INCLUDE_TAG = GitProtocolConstants.OPTION_INCLUDE_TAG;

    /** 
     * * Mutli-ACK support for improved negotiation.
     * * @since 2.0
     */
public static final String OPTION_MULTI_ACK = GitProtocolConstants.OPTION_MULTI_ACK;

    /** 
     * * Mutli-ACK detailed support for improved negotiation.
     * * @since 2.0
     */
public static final String OPTION_MULTI_ACK_DETAILED = GitProtocolConstants.OPTION_MULTI_ACK_DETAILED;

    /** 
     * * The client supports packs with deltas but not their bases.
     * * @since 2.0
     */
public static final String OPTION_THIN_PACK = GitProtocolConstants.OPTION_THIN_PACK;

    /** 
     * * The client supports using the side-band for progress messages.
     * * @since 2.0
     */
public static final String OPTION_SIDE_BAND = GitProtocolConstants.OPTION_SIDE_BAND;

    /** 
     * * The client supports using the 64K side-band for progress messages.
     * * @since 2.0
     */
public static final String OPTION_SIDE_BAND_64K = GitProtocolConstants.OPTION_SIDE_BAND_64K;

    /** 
     * * The client supports packs with OFS deltas.
     * * @since 2.0
     */
public static final String OPTION_OFS_DELTA = GitProtocolConstants.OPTION_OFS_DELTA;

    /** 
     * * The client supports shallow fetches.
     * * @since 2.0
     */
public static final String OPTION_SHALLOW = GitProtocolConstants.OPTION_SHALLOW;

    /** 
     * * The client does not want progress messages and will ignore them.
     * * @since 2.0
     */
public static final String OPTION_NO_PROGRESS = GitProtocolConstants.OPTION_NO_PROGRESS;

    /** 
     * * The client supports receiving a pack before it has sent "done".
     * * @since 2.0
     */
public static final String OPTION_NO_DONE = GitProtocolConstants.OPTION_NO_DONE;

    /** 
     * * The client supports fetching objects at the tip of any ref, even if not
     * * advertised.
     * * @since 3.1
     */
public static final String OPTION_ALLOW_TIP_SHA1_IN_WANT = GitProtocolConstants.OPTION_ALLOW_TIP_SHA1_IN_WANT;

    private final RevWalk walk;

    /** 
     * All commits that are immediately reachable by a local ref.
     */
private RevCommitList<org.eclipse.jgit.revwalk.RevCommit> reachableCommits;

    /** 
     * Marks an object as having all its dependencies.
     */
final RevFlag REACHABLE;

    /** 
     * Marks a commit known to both sides of the connection.
     */
final RevFlag COMMON;

    /** 
     * Like {@link #COMMON} but means its also in {@link #pckState}.
     */
private final RevFlag STATE;

    /** 
     * Marks a commit listed in the advertised refs.
     */
final RevFlag ADVERTISED;

    private GitProtocolConstants.MultiAck multiAck = GitProtocolConstants.MultiAck.OFF;

    private boolean thinPack;

    private boolean sideband;

    private boolean includeTags;

    private boolean allowOfsDelta;

    private boolean noDone;

    private boolean noProgress;

    private String lockMessage;

    private PackLock packLock;

    /** 
     * RPC state, if {@link BasePackConnection#statelessRPC} is true.
     */
private TemporaryBuffer.Heap state;

    private PacketLineOut pckState;

    /** 
     * * Create a new connection to fetch using the native git transport.
     * *
     * * @param packTransport
     * *            the transport.
     */
public BasePackFetchConnection(final PackTransport packTransport) {
        super(packTransport);
        if ((local) != null) {
            final FetchConfig cfg = local.getConfig().get(FetchConfig.KEY);
            allowOfsDelta = cfg.allowOfsDelta;
        } else {
            allowOfsDelta = true;
        }
        includeTags = (transport.getTagOpt()) != (TagOpt.NO_TAGS);
        thinPack = transport.isFetchThin();
        if ((local) != null) {
            walk = new RevWalk(local);
            reachableCommits = new RevCommitList<org.eclipse.jgit.revwalk.RevCommit>();
            REACHABLE = walk.newFlag("REACHABLE");
            COMMON = walk.newFlag("COMMON");
            STATE = walk.newFlag("STATE");
            ADVERTISED = walk.newFlag("ADVERTISED");
            walk.carry(COMMON);
            walk.carry(REACHABLE);
            walk.carry(ADVERTISED);
        } else {
            walk = null;
            REACHABLE = null;
            COMMON = null;
            STATE = null;
            ADVERTISED = null;
        }
    }

    private static class FetchConfig {
        static final Config.SectionParser<FetchConfig> KEY = new Config.SectionParser<FetchConfig>() {
            public FetchConfig parse(final Config cfg) {
                return new FetchConfig(cfg);
            }
        };

        final boolean allowOfsDelta;

        FetchConfig(final Config c) {
            allowOfsDelta = c.getBoolean("repack", "usedeltabaseoffset", true);
        }
    }

    public final void fetch(final ProgressMonitor monitor, final Collection<org.eclipse.jgit.lib.Ref> want, final Set<org.eclipse.jgit.lib.ObjectId> have) throws TransportException {
        fetch(monitor, want, have, null);
    }

    /** 
     * * @since 3.0
     */
public final void fetch(final ProgressMonitor monitor, final Collection<org.eclipse.jgit.lib.Ref> want, final Set<org.eclipse.jgit.lib.ObjectId> have, OutputStream outputStream) throws TransportException {
        markStartedOperation();
        doFetch(monitor, want, have, outputStream);
    }

    public boolean didFetchIncludeTags() {
        return false;
    }

    public boolean didFetchTestConnectivity() {
        return false;
    }

    public void setPackLockMessage(final String message) {
        lockMessage = message;
    }

    public Collection<org.eclipse.jgit.internal.storage.file.PackLock> getPackLocks() {
        if ((packLock) != null)
            return java.util.Collections.singleton(packLock);
        
        return java.util.Collections.<PackLock>emptyList();
    }

    /** 
     * * Execute common ancestor negotiation and fetch the objects.
     * *
     * * @param monitor
     * *            progress monitor to receive status updates. If the monitor is
     * *            the {@link NullProgressMonitor#INSTANCE}, then the no-progress
     * *            option enabled.
     * * @param want
     * *            the advertised remote references the caller wants to fetch.
     * * @param have
     * *            additional objects to assume that already exist locally. This
     * *            will be added to the set of objects reachable from the
     * *            destination repository's references.
     * * @param outputStream
     * *            ouputStream to write sideband messages to
     * * @throws TransportException
     * *             if any exception occurs.
     * * @since 3.0
     */
protected void doFetch(final ProgressMonitor monitor, final Collection<org.eclipse.jgit.lib.Ref> want, final Set<org.eclipse.jgit.lib.ObjectId> have, OutputStream outputStream) throws TransportException {
        try {
            noProgress = monitor == (NullProgressMonitor.INSTANCE);
            markRefsAdvertised();
            markReachable(have, maxTimeWanted(want));
            if (statelessRPC) {
                state = new TemporaryBuffer.Heap(Integer.MAX_VALUE);
                pckState = new PacketLineOut(state);
            } 
            if (sendWants(want)) {
                negotiate(monitor);
                walk.dispose();
                reachableCommits = null;
                state = null;
                pckState = null;
                receivePack(monitor, outputStream);
            } 
        } catch (CancelledException ce) {
            close();
            return ;
        } catch (IOException err) {
            close();
            throw new TransportException(err.getMessage() , err);
        } catch (RuntimeException err) {
            close();
            throw new TransportException(err.getMessage() , err);
        }
    }

    @Override
    public void close() {
        if ((walk) != null)
            walk.release();
        
        super.close();
    }

    private int maxTimeWanted(final Collection<org.eclipse.jgit.lib.Ref> wants) {
        int maxTime = 0;
        for (final Ref r : wants) {
            try {
                final RevObject obj = walk.parseAny(r.getObjectId());
                if (obj instanceof RevCommit) {
                    final int cTime = ((RevCommit)(obj)).getCommitTime();
                    if (maxTime < cTime)
                        maxTime = cTime;
                    
                } 
            } catch (IOException error) {
            }
        }
        return maxTime;
    }

    private void markReachable(final Set<org.eclipse.jgit.lib.ObjectId> have, final int maxTime) throws IOException {
        Map<java.lang.String, org.eclipse.jgit.lib.Ref> refs = local.getRefDatabase().getRefs(org.eclipse.jgit.lib.RefDatabase.ALL);
        for (final Ref r : refs.values()) {
            ObjectId id = r.getPeeledObjectId();
            if (id == null)
                id = r.getObjectId();
            
            if (id == null)
                continue;
            
            parseReachable(id);
        }
        for (ObjectId id : local.getAdditionalHaves())
            parseReachable(id);
        for (ObjectId id : have)
            parseReachable(id);
        if (maxTime > 0) {
            final Date maxWhen = new Date((maxTime * 1000L));
            walk.sort(org.eclipse.jgit.revwalk.RevSort.COMMIT_TIME_DESC);
            walk.markStart(reachableCommits);
            walk.setRevFilter(org.eclipse.jgit.revwalk.filter.CommitTimeRevFilter.after(maxWhen));
            for ( ;  ; ) {
                final RevCommit c = walk.next();
                if (c == null)
                    break;
                
                if ((c.has(ADVERTISED)) && (!(c.has(COMMON)))) {
                    c.add(COMMON);
                    c.carry(COMMON);
                    reachableCommits.add(c);
                } 
            }
        } 
    }

    private void parseReachable(ObjectId id) {
        try {
            RevCommit o = walk.parseCommit(id);
            if (!(o.has(REACHABLE))) {
                o.add(REACHABLE);
                reachableCommits.add(o);
            } 
        } catch (IOException readError) {
        }
    }

    private boolean sendWants(final Collection<org.eclipse.jgit.lib.Ref> want) throws IOException {
        final PacketLineOut p = statelessRPC ? pckState : pckOut;
        boolean first = true;
        for (final Ref r : want) {
            try {
                if (walk.parseAny(r.getObjectId()).has(REACHABLE)) {
                    continue;
                } 
            } catch (IOException err) {
            }
            final StringBuilder line = new StringBuilder(46);
            line.append("want ");
            line.append(r.getObjectId().name());
            if (first) {
                line.append(enableCapabilities());
                first = false;
            } 
            line.append('\n');
            p.writeString(line.toString());
        }
        if (first)
            return false;
        
        p.end();
        outNeedsEnd = false;
        return true;
    }

    private String enableCapabilities() throws TransportException {
        final StringBuilder line = new StringBuilder();
        if (noProgress)
            wantCapability(line, OPTION_NO_PROGRESS);
        
        if (includeTags)
            includeTags = wantCapability(line, OPTION_INCLUDE_TAG);
        
        if (allowOfsDelta)
            wantCapability(line, OPTION_OFS_DELTA);
        
        if (wantCapability(line, OPTION_MULTI_ACK_DETAILED)) {
            multiAck = GitProtocolConstants.MultiAck.DETAILED;
            if (statelessRPC)
                noDone = wantCapability(line, OPTION_NO_DONE);
            
        } else if (wantCapability(line, OPTION_MULTI_ACK))
            multiAck = GitProtocolConstants.MultiAck.CONTINUE;
        else
            multiAck = GitProtocolConstants.MultiAck.OFF;
        
        if (thinPack)
            thinPack = wantCapability(line, OPTION_THIN_PACK);
        
        if (wantCapability(line, OPTION_SIDE_BAND_64K))
            sideband = true;
        else if (wantCapability(line, OPTION_SIDE_BAND))
            sideband = true;
        
        if ((statelessRPC) && ((multiAck) != (GitProtocolConstants.MultiAck.DETAILED))) {
            throw new PackProtocolException(uri , java.text.MessageFormat.format(org.eclipse.jgit.internal.JGitText.get().statelessRPCRequiresOptionToBeEnabled, OPTION_MULTI_ACK_DETAILED));
        } 
        return line.toString();
    }

    private void negotiate(final ProgressMonitor monitor) throws IOException, CancelledException {
        final MutableObjectId ackId = new MutableObjectId();
        int resultsPending = 0;
        int havesSent = 0;
        int havesSinceLastContinue = 0;
        boolean receivedContinue = false;
        boolean receivedAck = false;
        boolean receivedReady = false;
        if (statelessRPC)
            state.writeTo(out, null);
        
        negotiateBegin();
        SEND_HAVES : for ( ;  ; ) {
            final RevCommit c = walk.next();
            if (c == null)
                break SEND_HAVES;
            
            pckOut.writeString((("have " + (c.getId().name())) + "\n"));
            havesSent++;
            havesSinceLastContinue++;
            if ((31 & havesSent) != 0) {
                continue;
            } 
            if (monitor.isCancelled())
                throw new CancelledException();
            
            pckOut.end();
            resultsPending++;
            if ((havesSent == 32) && (!(statelessRPC))) {
                continue;
            } 
            READ_RESULT : for ( ;  ; ) {
                final PacketLineIn.AckNackResult anr = pckIn.readACK(ackId);
                switch (anr) {
                    case NAK :
                        resultsPending--;
                        break READ_RESULT;
                    case ACK :
                        multiAck = GitProtocolConstants.MultiAck.OFF;
                        resultsPending = 0;
                        receivedAck = true;
                        if (statelessRPC)
                            state.writeTo(out, null);
                        
                        break SEND_HAVES;
                    case ACK_CONTINUE :
                    case ACK_COMMON :
                    case ACK_READY :
                        markCommon(walk.parseAny(ackId), anr);
                        receivedAck = true;
                        receivedContinue = true;
                        havesSinceLastContinue = 0;
                        if (anr == (PacketLineIn.AckNackResult.ACK_READY))
                            receivedReady = true;
                        
                        break;
                }
                if (monitor.isCancelled())
                    throw new CancelledException();
                
            }
            if ((noDone) & receivedReady)
                break SEND_HAVES;
            
            if (statelessRPC)
                state.writeTo(out, null);
            
            if (receivedContinue && (havesSinceLastContinue > (MAX_HAVES))) {
                break SEND_HAVES;
            } 
        }
        if (monitor.isCancelled())
            throw new CancelledException();
        
        if ((!receivedReady) || (!(noDone))) {
            pckOut.writeString("done\n");
            pckOut.flush();
        } 
        if (!receivedAck) {
            multiAck = GitProtocolConstants.MultiAck.OFF;
            resultsPending++;
        } 
        READ_RESULT : while ((resultsPending > 0) || ((multiAck) != (GitProtocolConstants.MultiAck.OFF))) {
            final PacketLineIn.AckNackResult anr = pckIn.readACK(ackId);
            resultsPending--;
            switch (anr) {
                case NAK :
                    break;
                case ACK :
                    break READ_RESULT;
                case ACK_CONTINUE :
                case ACK_COMMON :
                case ACK_READY :
                    multiAck = GitProtocolConstants.MultiAck.CONTINUE;
                    break;
            }
            if (true) {
                if (monitor.isCancelled())
                    throw new CancelledException();
                
                monitor.endTask();
            } 
        }
    }

    private void negotiateBegin() throws IOException {
        walk.resetRetain(REACHABLE, ADVERTISED);
        walk.markStart(reachableCommits);
        walk.sort(org.eclipse.jgit.revwalk.RevSort.COMMIT_TIME_DESC);
        walk.setRevFilter(new org.eclipse.jgit.revwalk.filter.RevFilter() {
            @Override
            public org.eclipse.jgit.revwalk.filter.RevFilter clone() {
                return this;
            }

            @Override
            public boolean include(final RevWalk walker, final RevCommit c) {
                final boolean remoteKnowsIsCommon = c.has(COMMON);
                if (c.has(ADVERTISED)) {
                    c.add(COMMON);
                } 
                return !remoteKnowsIsCommon;
            }

            @Override
            public boolean requiresCommitBody() {
                return false;
            }
        });
    }

    private void markRefsAdvertised() {
        for (final Ref r : getRefs()) {
            markAdvertised(r.getObjectId());
            if ((r.getPeeledObjectId()) != null)
                markAdvertised(r.getPeeledObjectId());
            
        }
    }

    private void markAdvertised(final AnyObjectId id) {
        try {
            walk.parseAny(id).add(ADVERTISED);
        } catch (IOException readError) {
        }
    }

    private void markCommon(final RevObject obj, final PacketLineIn.AckNackResult anr) throws IOException {
        if (((statelessRPC) && (anr == (PacketLineIn.AckNackResult.ACK_COMMON))) && (!(obj.has(STATE)))) {
            StringBuilder s;
            s = new StringBuilder((6 + (Constants.OBJECT_ID_STRING_LENGTH)));
            s.append("have ");
            s.append(obj.name());
            s.append('\n');
            pckState.writeString(s.toString());
            obj.add(STATE);
        } 
        obj.add(COMMON);
        if (obj instanceof RevCommit)
            ((RevCommit)(obj)).carry(COMMON);
        
    }

    private void receivePack(final ProgressMonitor monitor, OutputStream outputStream) throws IOException {
        onReceivePack();
        InputStream input = in;
        if (sideband)
            input = new SideBandInputStream(input , monitor , getMessageWriter() , outputStream);
        
        ObjectInserter ins = local.newObjectInserter();
        try {
            PackParser parser = ins.newPackParser(input);
            parser.setAllowThin(thinPack);
            parser.setObjectChecker(transport.getObjectChecker());
            parser.setLockMessage(lockMessage);
            packLock = parser.parse(monitor);
            ins.flush();
        } finally {
            ins.release();
        }
    }

    /** 
     * * Notification event delivered just before the pack is received from the
     * * network. This event can be used by RPC such as {@link TransportHttp} to
     * * disable its request magic and ensure the pack stream is read correctly.
     * *
     * * @since 2.0
     */
protected void onReceivePack() {
    }

    private static class CancelledException extends Exception {
        private static final long serialVersionUID = 1L;
    }
}

