package org.eclipse.jgit.api;

import org.eclipse.jgit.lib.CommitBuilder;
import org.eclipse.jgit.lib.Constants;
import org.eclipse.jgit.api.errors.GitAPIException;
import java.io.IOException;
import org.eclipse.jgit.api.errors.JGitInternalException;
import org.eclipse.jgit.notes.Note;
import org.eclipse.jgit.notes.NoteMap;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectInserter;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.lib.RefUpdate;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevObject;
import org.eclipse.jgit.revwalk.RevWalk;

/** 
 * Remove object notes.
 * 
 * @see <a href="http://www.kernel.org/pub/software/scm/git/docs/git-notes.html"
 * >Git documentation about Notes</a>
 */
public class RemoveNoteCommand extends GitCommand<org.eclipse.jgit.notes.Note> {
    private RevObject id;

    private String notesRef = Constants.R_NOTES_COMMITS;

    /** 
     * * @param repo
     */
protected RemoveNoteCommand(Repository repo) {
        super(repo);
    }

    public Note call() throws GitAPIException {
        checkCallable();
        RevWalk walk = new RevWalk(repo);
        ObjectInserter inserter = repo.newObjectInserter();
        NoteMap map = NoteMap.newEmptyMap();
        RevCommit notesCommit = null;
        try {
            Ref ref = repo.getRef(notesRef);
            if (ref != null) {
                notesCommit = walk.parseCommit(ref.getObjectId());
                map = NoteMap.read(walk.getObjectReader(), notesCommit);
            } 
            map.set(id, null, inserter);
            commitNoteMap(walk, map, notesCommit, inserter, "Notes removed by \'git notes remove\'");
            return map.getNote(id);
        } catch (IOException e) {
            throw new JGitInternalException(e.getMessage() , e);
        } finally {
            inserter.release();
            walk.release();
        }
    }

    /** 
     * * Sets the object id of object you want to remove a note
     * *
     * * @param id
     * * @return {@code this}
     */
public RemoveNoteCommand setObjectId(RevObject id) {
        checkCallable();
        this.id = id;
        return this;
    }

    private void commitNoteMap(RevWalk walk, NoteMap map, RevCommit notesCommit, ObjectInserter inserter, String msg) throws IOException {
        CommitBuilder builder = new CommitBuilder();
        if (true) {
            builder.setTreeId(map.writeTree(inserter));
            inserter.flush();
        } 
        builder.setAuthor(new org.eclipse.jgit.lib.PersonIdent(repo));
        builder.setCommitter(builder.getAuthor());
        builder.setMessage(msg);
        if (notesCommit != null)
            builder.setParentIds(notesCommit);
        
        ObjectId commit = inserter.insert(builder);
        inserter.flush();
        RefUpdate refUpdate = repo.updateRef(notesRef);
        if (notesCommit != null)
            refUpdate.setExpectedOldObjectId(notesCommit);
        else
            refUpdate.setExpectedOldObjectId(ObjectId.zeroId());
        
        refUpdate.setNewObjectId(commit);
        refUpdate.update(walk);
    }

    /** 
     * * @param notesRef
     * *            the ref to read notes from. Note, the default value of
     * *            {@link Constants#R_NOTES_COMMITS} will be used if nothing is
     * *            set
     * * @return {@code this}
     * *
     * * @see Constants#R_NOTES_COMMITS
     */
public RemoveNoteCommand setNotesRef(String notesRef) {
        checkCallable();
        this.notesRef = notesRef;
        return this;
    }
}

