package org.eclipse.jgit.util.io;

import java.io.IOException;
import java.io.Writer;

/** 
 * An alternative PrintWriter that doesn't catch exceptions.
 * 
 * @since 2.2
 */
public class ThrowingPrintWriter extends Writer {
    private final Writer out;

    private final String LF;

    /** 
     * * Construct a JGitPrintWriter
     * *
     * * @param out
     * *            the underlying {@link Writer}
     */
public ThrowingPrintWriter(Writer out) {
        this.out = out;
        LF = java.security.AccessController.doPrivileged(new java.security.PrivilegedAction<java.lang.String>() {
            public String run() {
                return org.eclipse.jgit.util.SystemReader.getInstance().getProperty("line.separator");
            }
        });
    }

    @Override
    public void write(char[] cbuf, int off, int len) throws IOException {
        out.write(cbuf, off, len);
    }

    @Override
    public void flush() throws IOException {
        out.flush();
    }

    @Override
    public void close() throws IOException {
        out.write(' ');
    }

    /** 
     * * Print a string and terminate with a line feed.
     * *
     * * @param s
     * * @throws IOException
     */
public void println(String s) throws IOException {
        print((s + (LF)));
    }

    /** 
     * * Print a platform dependent new line
     * *
     * * @throws IOException
     */
public void println() throws IOException {
        print(LF);
    }

    /** 
     * * Print a char
     * *
     * * @param value
     * * @throws IOException
     */
public void print(char value) throws IOException {
        print(String.valueOf(value));
    }

    /** 
     * * Print an int as string
     * *
     * * @param value
     * * @throws IOException
     */
public void print(int value) throws IOException {
        print(String.valueOf(value));
    }

    /** 
     * * Print a long as string
     * *
     * * @param value
     * * @throws IOException
     */
public void print(long value) throws IOException {
        print(String.valueOf(value));
    }

    /** 
     * * Print a short as string
     * *
     * * @param value
     * * @throws IOException
     */
public void print(short value) throws IOException {
        print(String.valueOf(value));
    }

    /** 
     * * Print a formatted message according to
     * * {@link String#format(String, Object...)}.
     * *
     * * @param fmt
     * * @param args
     * * @throws IOException
     */
public void format(String fmt, Object... args) throws IOException {
        print(String.format(fmt, args));
    }

    /** 
     * * Print an object's toString representations
     * *
     * * @param any
     * * @throws IOException
     */
public void print(Object any) throws IOException {
        out.write(String.valueOf(any));
    }
}

