package org.eclipse.jgit.api;

import java.io.File;
import java.io.IOException;
import org.eclipse.jgit.api.errors.JGitInternalException;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.junit.RepositoryTestCase;
import org.junit.Test;
import org.eclipse.jgit.treewalk.TreeWalk;

/** 
 * Testing the 'commit only' option:
 * 
 * I. A single file (f1.txt) specified as part of the --only/ -o option can have
 * one of the following (14) states:
 * 
 * <pre>
 * |                          | expected result
 * ---------------------------------------------------------------------
 * | HEAD  DirCache  Worktree | HEAD  DirCache
 * ---------------------------------------------------------------------
 * f1_1  |  -       -       c       |                => e: path unknown
 * f1_2  |  -       c       -       |                => no changes
 * f1_3  |  c       -       -       |  -       -
 * f1_4  |  -       c       c       |  c       c
 * f1_5  |  c       c       -       |  -       -
 * f1_6  |  c       -       c       |                => no changes
 * f1_7  |  c       c       c       |                => no changes
 * ---------------------------------------------------------------------
 * f1_8  |  -       c       c'      |  c'      c'
 * f1_9  |  c       -       c'      |  c'      c'
 * f1_10  |  c       c'      -       |  -       -
 * f1_11  |  c       c       c'      |  c'      c'
 * f1_12  |  c       c'      c       |                => no changes
 * f1_13  |  c       c'      c'      |  c'      c'
 * ---------------------------------------------------------------------
 * f1_14  |  c       c'      c''     |  c''     c''
 * </pre>
 * 
 * II. Scenarios that do not end with a successful commit (1, 2, 6, 7, 12) have
 * to be tested with a second file (f2.txt) specified that would lead to a
 * successful commit, if it were executed separately (e.g. scenario 14).
 * 
 * <pre>
 * |                          | expected result
 * ---------------------------------------------------------------------------
 * | HEAD  DirCache  Worktree | HEAD  DirCache
 * ---------------------------------------------------------------------------
 * f1_1_f2_14  |  -       -       c       |                => e: path unknown
 * f1_2_f2_14  |  -       c       -       |  -       -
 * f1_6_f2_14  |  c       -       c       |  c       c
 * f1_7_f2_14  |  c       c       c       |  c       c
 * ---------------------------------------------------------------------------
 * f1_12_f2_14  |  c       c'      c       |  c       c
 * </pre>
 * 
 * III. All scenarios (1-14, I-II) have to be tested with different repository
 * states, to check that the --only/ -o option does not change existing content
 * (HEAD and DirCache). The following states for a file (f3.txt) not specified
 * shall be tested:
 * 
 * <pre>
 * | HEAD  DirCache
 * --------------------
 * *_a  |  -       -
 * *_b  |  -       c
 * *_c  |  c       c
 * *_d  |  c       -
 * --------------------
 * *_e  |  c       c'
 * </pre>
 */
public class CommitOnlyTest extends RepositoryTestCase {
    private static final String F1 = "f1.txt";

    private static final String F2 = "f2.txt";

    private static final String F3 = "f3.txt";

    private static final String MSG = "commit";

    private static int A = 0;

    private static int B = 1;

    private static int C = 2;

    private static int D = 3;

    private static int E = 4;

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a_add380() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a_add380");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a_add381() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a_add381");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a_remove375() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a_remove375");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a_remove376() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a_remove376");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_1(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_a_remove377() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_a_remove377");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),630,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b_add383() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b_add383");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b_add384() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b_add384");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b_remove378() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b_remove378");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b_remove379() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b_remove379");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_1(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_b_remove380() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_b_remove380");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),631,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c_add386() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c_add386");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c_add387() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c_add387");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c_remove381() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c_remove381");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c_remove382() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c_remove382");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_1(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_c_remove383() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_c_remove383");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),632,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d_add389() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d_add389");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d_add390() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d_add390");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d_remove384() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d_remove384");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d_remove385() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d_remove385");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_1(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_d_remove386() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_d_remove386");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),277,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e_add392() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e_add392");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e_add393() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e_add393");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e_remove387() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e_remove387");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e_remove388() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e_remove388");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_1(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_e_remove389() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_e_remove389");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),633,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a_add395() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a_add395");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a_add396() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a_add396");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a_remove390() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a_remove390");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a_remove391() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a_remove391");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_a_remove392() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_a_remove392");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),634,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b_add398() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b_add398");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b_add399() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b_add399");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b_remove393() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b_remove393");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b_remove394() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b_remove394");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_b_remove395() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_b_remove395");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),635,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c_add401() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c_add401");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c_add402() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c_add402");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c_remove396() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c_remove396");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c_remove397() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c_remove397");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_c_remove398() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_c_remove398");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),11,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d_add404() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d_add404");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d_add405() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d_add405");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d_remove399() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d_remove399");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d_remove400() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d_remove400");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_d_remove401() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_d_remove401");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),636,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e_add407() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e_add407");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e_add408() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e_add408");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e_remove402() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e_remove402");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f1_1(git);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e_remove403() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e_remove403");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_1_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_1_f2_14_e_remove404() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_1_f2_14_e_remove404");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),637,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_1(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a_add410() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a_add410");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a_add411() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a_add411");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a_remove405() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a_remove405");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a_remove406() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a_remove406");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_2(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_a_remove407() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_a_remove407");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),638,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b_add413() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b_add413");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b_add414() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b_add414");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b_remove408() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b_remove408");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b_remove409() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b_remove409");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_2(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_b_remove410() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_b_remove410");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),639,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c_add416() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c_add416");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c_add417() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c_add417");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c_remove411() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c_remove411");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c_remove412() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c_remove412");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_2(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_c_remove413() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_c_remove413");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),640,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d_add419() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d_add419");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d_add420() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d_add420");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d_remove414() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d_remove414");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d_remove415() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d_remove415");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_2(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_d_remove416() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_d_remove416");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),641,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e_add422() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e_add422");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e_add423() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e_add423");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e_remove417() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e_remove417");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e_remove418() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e_remove418");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_2(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_e_remove419() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_e_remove419");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),642,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a_add425() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a_add425");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a_add426() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a_add426");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a_remove420() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a_remove420");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a_remove421() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a_remove421");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_a_remove422() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_a_remove422");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),643,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b_add428() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b_add428");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b_add429() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b_add429");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b_remove423() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b_remove423");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b_remove424() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b_remove424");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_b_remove425() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_b_remove425");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),644,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c_add431() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c_add431");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c_add432() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c_add432");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c_remove426() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c_remove426");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c_remove427() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c_remove427");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_c_remove428() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_c_remove428");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),645,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d_add434() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d_add434");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d_add435() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d_add435");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d_remove429() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d_remove429");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d_remove430() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d_remove430");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_d_remove431() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_d_remove431");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),646,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e_add437() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e_add437");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e_add438() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e_add438");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e_remove432() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e_remove432");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f1_2(git);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e_remove433() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e_remove433");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_2_f2_f14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_2_f2_14_e_remove434() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_2_f2_14_e_remove434");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),647,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_2(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a_add440() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a_add440");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_3(git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a_add441() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a_add441");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a_remove435() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a_remove435");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a_remove436() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a_remove436");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_3(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_a_remove437() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_a_remove437");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),648,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_3(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b_add443() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b_add443");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_3(git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b_add444() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b_add444");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b_remove438() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b_remove438");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b_remove439() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b_remove439");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_3(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_b_remove440() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_b_remove440");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),649,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_3(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c_add446() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c_add446");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_3(git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c_add447() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c_add447");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c_remove441() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c_remove441");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c_remove442() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c_remove442");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_3(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_c_remove443() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_c_remove443");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),650,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_3(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d_add449() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d_add449");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_3(git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d_add450() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d_add450");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d_remove444() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d_remove444");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d_remove445() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d_remove445");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_3(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_d_remove446() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_d_remove446");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),651,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_3(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e_add452() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e_add452");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_3(git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e_add453() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e_add453");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e_remove447() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e_remove447");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f1_3(git);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e_remove448() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e_remove448");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_3(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_3_e_remove449() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_3_e_remove449");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),652,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_3(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a_add455() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a_add455");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_4(git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a_add456() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a_add456");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a_remove450() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a_remove450");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a_remove451() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a_remove451");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_4(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_a_remove452() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_a_remove452");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),653,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_4(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b_add458() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b_add458");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_4(git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b_add459() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b_add459");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b_remove453() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b_remove453");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b_remove454() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b_remove454");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_4(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_b_remove455() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_b_remove455");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),654,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_4(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c_add461() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c_add461");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_4(git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c_add462() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c_add462");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c_remove456() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c_remove456");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c_remove457() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c_remove457");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_4(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_c_remove458() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_c_remove458");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),655,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_4(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d_add464() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d_add464");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_4(git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d_add465() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d_add465");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d_remove459() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d_remove459");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d_remove460() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d_remove460");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_4(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_d_remove461() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_d_remove461");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),656,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_4(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e_add467() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e_add467");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_4(git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e_add468() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e_add468");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e_remove462() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e_remove462");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f1_4(git);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e_remove463() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e_remove463");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_4(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_4_e_remove464() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_4_e_remove464");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),657,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_4(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a_add470() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a_add470");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_5(git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a_add471() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a_add471");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a_remove465() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a_remove465");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a_remove466() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a_remove466");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_5(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_a_remove467() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_a_remove467");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),658,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_5(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b_add473() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b_add473");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_5(git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b_add474() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b_add474");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b_remove468() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b_remove468");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b_remove469() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b_remove469");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_5(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_b_remove470() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_b_remove470");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),659,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_5(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c_add476() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c_add476");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_5(git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c_add477() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c_add477");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c_remove471() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c_remove471");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c_remove472() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c_remove472");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_5(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_c_remove473() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_c_remove473");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),660,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_5(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d_add479() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d_add479");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_5(git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d_add480() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d_add480");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d_remove474() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d_remove474");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d_remove475() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d_remove475");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_5(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_d_remove476() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_d_remove476");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),661,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_5(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e_add482() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e_add482");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_5(git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e_add483() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e_add483");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e_remove477() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e_remove477");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f1_5(git);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e_remove478() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e_remove478");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_5(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_5_e_remove479() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_5_e_remove479");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),662,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_5(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a_add485() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a_add485");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a_add486() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a_add486");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a_remove480() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a_remove480");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a_remove481() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a_remove481");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_6(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_a_remove482() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_a_remove482");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),663,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b_add488() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b_add488");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b_add489() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b_add489");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b_remove483() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b_remove483");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b_remove484() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b_remove484");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_6(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_b_remove485() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_b_remove485");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),664,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c_add491() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c_add491");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c_add492() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c_add492");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c_remove486() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c_remove486");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c_remove487() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c_remove487");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_6(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_c_remove488() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_c_remove488");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),665,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d_add494() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d_add494");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d_add495() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d_add495");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d_remove489() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d_remove489");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d_remove490() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d_remove490");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_6(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_d_remove491() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_d_remove491");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),666,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e_add497() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e_add497");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e_add498() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e_add498");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e_remove492() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e_remove492");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e_remove493() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e_remove493");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_6(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_e_remove494() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_e_remove494");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),667,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a_add500() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a_add500");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a_add501() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a_add501");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a_remove495() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a_remove495");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a_remove496() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a_remove496");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_a_remove497() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_a_remove497");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),668,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b_add503() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b_add503");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b_add504() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b_add504");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b_remove498() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b_remove498");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b_remove499() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b_remove499");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_b_remove500() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_b_remove500");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),669,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c_add506() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c_add506");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c_add507() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c_add507");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c_remove501() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c_remove501");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c_remove502() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c_remove502");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_c_remove503() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_c_remove503");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),670,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d_add509() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d_add509");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d_add510() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d_add510");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d_remove504() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d_remove504");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d_remove505() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d_remove505");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_d_remove506() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_d_remove506");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),671,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e_add512() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e_add512");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e_add513() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e_add513");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e_remove507() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e_remove507");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f1_6(git);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e_remove508() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e_remove508");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_6_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_6_f2_14_e_remove509() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_6_f2_14_e_remove509");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),672,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_6(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a_add515() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a_add515");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a_add516() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a_add516");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a_remove510() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a_remove510");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a_remove511() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a_remove511");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_7(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_a_remove512() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_a_remove512");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),673,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b_add518() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b_add518");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b_add519() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b_add519");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b_remove513() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b_remove513");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b_remove514() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b_remove514");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_7(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_b_remove515() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_b_remove515");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),674,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c_add521() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c_add521");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c_add522() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c_add522");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c_remove516() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c_remove516");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c_remove517() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c_remove517");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_7(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_c_remove518() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_c_remove518");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),675,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d_add524() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d_add524");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d_add525() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d_add525");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d_remove519() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d_remove519");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d_remove520() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d_remove520");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_7(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_d_remove521() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_d_remove521");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),676,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e_add527() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e_add527");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e_add528() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e_add528");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e_remove522() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e_remove522");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e_remove523() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e_remove523");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_7(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_e_remove524() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_e_remove524");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),677,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a_add530() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a_add530");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a_add531() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a_add531");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a_remove525() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a_remove525");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a_remove526() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a_remove526");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_a_remove527() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_a_remove527");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),678,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b_add533() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b_add533");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b_add534() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b_add534");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b_remove528() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b_remove528");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b_remove529() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b_remove529");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_b_remove530() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_b_remove530");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),679,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c_add536() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c_add536");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c_add537() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c_add537");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c_remove531() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c_remove531");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c_remove532() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c_remove532");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_c_remove533() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_c_remove533");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),680,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d_add539() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d_add539");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d_add540() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d_add540");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d_remove534() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d_remove534");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d_remove535() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d_remove535");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_d_remove536() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_d_remove536");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),681,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e_add542() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e_add542");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e_add543() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e_add543");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e_remove537() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e_remove537");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f1_7(git);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e_remove538() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e_remove538");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_7_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_7_f2_14_e_remove539() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_7_f2_14_e_remove539");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),682,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_7(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a_add545() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a_add545");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_8(git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a_add546() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a_add546");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a_remove540() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a_remove540");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a_remove541() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a_remove541");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_8(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_a_remove542() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_a_remove542");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),683,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_8(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b_add548() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b_add548");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_8(git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b_add549() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b_add549");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b_remove543() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b_remove543");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b_remove544() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b_remove544");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_8(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_b_remove545() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_b_remove545");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),684,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_8(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c_add551() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c_add551");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_8(git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c_add552() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c_add552");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c_remove546() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c_remove546");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c_remove547() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c_remove547");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_8(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_c_remove548() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_c_remove548");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),685,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_8(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d_add554() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d_add554");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_8(git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d_add555() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d_add555");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d_remove549() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d_remove549");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d_remove550() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d_remove550");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_8(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_d_remove551() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_d_remove551");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),686,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_8(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e_add557() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e_add557");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_8(git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e_add558() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e_add558");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e_remove552() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e_remove552");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f1_8(git);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e_remove553() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e_remove553");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_8(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_8_e_remove554() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_8_e_remove554");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),687,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_8(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a_add560() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a_add560");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_9(git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a_add561() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a_add561");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a_remove555() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a_remove555");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a_remove556() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a_remove556");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_9(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_a_remove557() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_a_remove557");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),688,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_9(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b_add563() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b_add563");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_9(git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b_add564() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b_add564");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b_remove558() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b_remove558");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b_remove559() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b_remove559");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_9(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_b_remove560() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_b_remove560");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),689,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_9(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c_add566() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c_add566");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_9(git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c_add567() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c_add567");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c_remove561() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c_remove561");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c_remove562() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c_remove562");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_9(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_c_remove563() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_c_remove563");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),150,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_9(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d_add569() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d_add569");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_9(git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d_add570() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d_add570");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d_remove564() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d_remove564");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d_remove565() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d_remove565");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_9(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_d_remove566() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_d_remove566");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),690,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_9(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e_add572() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e_add572");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_9(git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e_add573() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e_add573");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e_remove567() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e_remove567");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f1_9(git);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e_remove568() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e_remove568");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_9(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_9_e_remove569() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_9_e_remove569");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),691,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_9(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a_add290() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a_add290");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_10(git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a_add291() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a_add291");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a_remove285() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a_remove285");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a_remove286() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a_remove286");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_10(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_a_remove287() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_a_remove287");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),600,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_10(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b_add293() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b_add293");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_10(git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b_add294() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b_add294");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b_remove288() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b_remove288");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b_remove289() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b_remove289");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_10(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_b_remove290() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_b_remove290");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),601,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_10(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c_add296() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c_add296");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_10(git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c_add297() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c_add297");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c_remove291() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c_remove291");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c_remove292() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c_remove292");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_10(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_c_remove293() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_c_remove293");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),602,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_10(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d_add299() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d_add299");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_10(git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d_add300() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d_add300");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d_remove294() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d_remove294");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d_remove295() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d_remove295");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_10(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_d_remove296() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_d_remove296");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),603,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_10(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e_add302() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e_add302");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_10(git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e_add303() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e_add303");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e_remove297() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e_remove297");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f1_10(git);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e_remove298() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e_remove298");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_10(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_10_e_remove299() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_10_e_remove299");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),604,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_10(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a_add305() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a_add305");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_11(git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a_add306() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a_add306");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a_remove300() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a_remove300");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a_remove301() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a_remove301");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_11(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_a_remove302() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_a_remove302");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),605,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_11(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b_add308() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b_add308");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_11(git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b_add309() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b_add309");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b_remove303() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b_remove303");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b_remove304() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b_remove304");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_11(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_b_remove305() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_b_remove305");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),606,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_11(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c_add311() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c_add311");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_11(git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c_add312() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c_add312");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c_remove306() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c_remove306");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c_remove307() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c_remove307");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_11(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_c_remove308() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_c_remove308");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),607,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_11(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d_add314() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d_add314");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_11(git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d_add315() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d_add315");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d_remove309() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d_remove309");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d_remove310() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d_remove310");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_11(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_d_remove311() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_d_remove311");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),608,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_11(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e_add317() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e_add317");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_11(git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e_add318() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e_add318");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e_remove312() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e_remove312");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f1_11(git);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e_remove313() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e_remove313");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_11(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_11_e_remove314() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_11_e_remove314");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),609,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_11(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a_add320() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a_add320");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a_add321() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a_add321");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a_remove315() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a_remove315");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a_remove316() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a_remove316");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_12(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_a_remove317() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_a_remove317");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),610,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b_add323() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b_add323");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b_add324() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b_add324");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b_remove318() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b_remove318");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b_remove319() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b_remove319");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_12(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_b_remove320() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_b_remove320");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),611,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c_add326() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c_add326");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c_add327() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c_add327");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c_remove321() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c_remove321");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c_remove322() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c_remove322");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_12(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_c_remove323() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_c_remove323");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),612,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d_add329() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d_add329");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d_add330() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d_add330");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d_remove324() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d_remove324");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d_remove325() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d_remove325");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_12(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_d_remove326() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_d_remove326");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),613,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e_add332() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e_add332");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e_add333() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e_add333");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e_remove327() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e_remove327");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e_remove328() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e_remove328");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_12(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_e_remove329() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_e_remove329");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),614,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a_add335() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a_add335");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a_add336() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a_add336");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a_remove330() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a_remove330");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a_remove331() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a_remove331");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_a_remove332() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_a_remove332");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),615,git);
            prepare_f3_f2_14(git, CommitOnlyTest.A);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b_add338() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b_add338");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b_add339() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b_add339");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b_remove333() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b_remove333");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b_remove334() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b_remove334");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_b_remove335() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_b_remove335");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),616,git);
            prepare_f3_f2_14(git, CommitOnlyTest.B);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c_add341() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c_add341");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c_add342() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c_add342");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c_remove336() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c_remove336");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c_remove337() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c_remove337");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_c_remove338() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_c_remove338");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),617,git);
            prepare_f3_f2_14(git, CommitOnlyTest.C);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d_add344() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d_add344");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d_add345() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d_add345");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d_remove339() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d_remove339");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d_remove340() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d_remove340");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_d_remove341() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_d_remove341");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),618,git);
            prepare_f3_f2_14(git, CommitOnlyTest.D);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e_add347() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e_add347");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e_add348() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e_add348");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e_remove342() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e_remove342");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f1_12(git);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e_remove343() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e_remove343");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_12_f2_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_12_f2_14_e_remove344() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_12_f2_14_e_remove344");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),619,git);
            prepare_f3_f2_14(git, CommitOnlyTest.E);
            prepare_f1_12(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a_add350() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a_add350");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_13(git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a_add351() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a_add351");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a_remove345() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a_remove345");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a_remove346() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a_remove346");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_13(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_a_remove347() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_a_remove347");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),620,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_13(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b_add353() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b_add353");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_13(git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b_add354() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b_add354");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b_remove348() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b_remove348");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b_remove349() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b_remove349");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_13(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_b_remove350() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_b_remove350");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),621,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_13(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c_add356() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c_add356");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_13(git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c_add357() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c_add357");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c_remove351() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c_remove351");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c_remove352() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c_remove352");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_13(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_c_remove353() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_c_remove353");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),622,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_13(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d_add359() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d_add359");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_13(git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d_add360() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d_add360");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d_remove354() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d_remove354");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d_remove355() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d_remove355");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_13(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_d_remove356() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_d_remove356");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),623,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_13(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e_add362() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e_add362");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_13(git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e_add363() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e_add363");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e_remove357() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e_remove357");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f1_13(git);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e_remove358() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e_remove358");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_13(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_13_e_remove359() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_13_e_remove359");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),624,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_13(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a_add365() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a_add365");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_14(git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a_add366() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a_add366");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a_remove360() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a_remove360");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a_remove361() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a_remove361");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f3(git, CommitOnlyTest.A);
            executeAndCheck_f1_14(git, CommitOnlyTest.A);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_a_remove362() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_a_remove362");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),625,git);
            prepare_f3(git, CommitOnlyTest.A);
            prepare_f1_14(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b_add368() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b_add368");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_14(git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b_add369() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b_add369");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b_remove363() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b_remove363");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b_remove364() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b_remove364");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f3(git, CommitOnlyTest.B);
            executeAndCheck_f1_14(git, CommitOnlyTest.B);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_b_remove365() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_b_remove365");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),626,git);
            prepare_f3(git, CommitOnlyTest.B);
            prepare_f1_14(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c_add371() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c_add371");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_14(git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c_add372() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c_add372");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c_remove366() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c_remove366");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c_remove367() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c_remove367");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f3(git, CommitOnlyTest.C);
            executeAndCheck_f1_14(git, CommitOnlyTest.C);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_c_remove368() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_c_remove368");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),627,git);
            prepare_f3(git, CommitOnlyTest.C);
            prepare_f1_14(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d_add374() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d_add374");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_14(git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d_add375() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d_add375");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d_remove369() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d_remove369");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d_remove370() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d_remove370");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f3(git, CommitOnlyTest.D);
            executeAndCheck_f1_14(git, CommitOnlyTest.D);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_d_remove371() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_d_remove371");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),628,git);
            prepare_f3(git, CommitOnlyTest.D);
            prepare_f1_14(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e_add377() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e_add377");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_14(git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e_add378() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e_add378");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e_remove372() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e_remove372");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f1_14(git);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e_remove373() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e_remove373");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f3(git, CommitOnlyTest.E);
            executeAndCheck_f1_14(git, CommitOnlyTest.E);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOption_f1_14_e_remove374() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOption_f1_14_e_remove374");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),629,git);
            prepare_f3(git, CommitOnlyTest.E);
            prepare_f1_14(git);
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add275() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add275");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add276() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add276");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add277() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add277");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add278() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add278");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add279() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add279");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add280() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add280");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add281() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add281");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add282() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add282");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add283() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add283");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add284() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add284");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add285() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add285");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add286() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add286");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add287() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add287");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_add288() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_add288");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test
    public void testOnlyOptionWithDirectory() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test
    public void testOnlyOptionWithDirectory_literalMutation994() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_literalMutation994");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/dN2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test
    public void testOnlyOptionWithDirectory_literalMutation995() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_literalMutation995");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove271() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove271");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove272() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove272");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove273() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove273");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove274() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove274");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove275() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove275");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove276() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove276");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove277() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove277");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove278() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove278");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove279() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove279");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove280() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove280");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove281() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove281");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove282() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove282");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove283() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove283");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            git.commit().setOnly("d1").setOnly("d3/d4/").setOnly("d5").setOnly("d6/f11.txt").setMessage(MSG).call();
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @Test(timeout = 1000)
    public void testOnlyOptionWithDirectory_remove284() throws Exception {
        try {
            fr.inria.diversify.testamplification.logger.Logger.writeTestStart(Thread.currentThread(),this, "testOnlyOptionWithDirectory_remove284");
            final Git git = new Git(db);
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),593,git);
            final File f1 = writeTrashFile("d1/d2/f1.txt", "c1");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),594,f1);
            writeTrashFile("d1/d2/f2.txt", "c2");
            final File f3 = writeTrashFile("d1/f3.txt", "c3");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),595,f3);
            writeTrashFile("d1/f4.txt", "c4");
            final File f5 = writeTrashFile("d3/d4/f5.txt", "c5");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),596,f5);
            writeTrashFile("d3/d4/f6.txt", "c6");
            final File f7 = writeTrashFile("d3/f7.txt", "c7");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),597,f7);
            writeTrashFile("d3/f8.txt", "c8");
            final File f9 = writeTrashFile("d5/f9.txt", "c9");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),598,f9);
            writeTrashFile("d5/f10.txt", "c10");
            final File f11 = writeTrashFile("d6/f11.txt", "c11");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),599,f11);
            writeTrashFile("d6/f12.txt", "c12");
            git.add().addFilepattern(".").call();
            write(f1, "c1\'");
            write(f3, "c3\'");
            write(f5, "c5\'");
            write(f7, "c7\'");
            write(f9, "c9\'");
            write(f11, "c11\'");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),579,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f1.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),580,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/d2/f2.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),581,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f3.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),582,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d1/f4.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),583,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f5.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),584,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/d4/f6.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),585,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f7.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),586,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d3/f8.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),587,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f9.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),588,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d5/f10.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),589,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f11.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),590,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, "d6/f12.txt"));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),591,("[d1/d2/f1.txt, mode:100644, content:c1\']" + ("[d1/d2/f2.txt, mode:100644, content:c2]" + ("[d1/f3.txt, mode:100644, content:c3\']" + ("[d1/f4.txt, mode:100644, content:c4]" + ("[d3/d4/f5.txt, mode:100644, content:c5\']" + ("[d3/d4/f6.txt, mode:100644, content:c6]" + ("[d3/f7.txt, mode:100644, content:c7]" + ("[d3/f8.txt, mode:100644, content:c8]" + ("[d5/f10.txt, mode:100644, content:c10]" + ("[d5/f9.txt, mode:100644, content:c9\']" + ("[d6/f11.txt, mode:100644, content:c11\']" + "[d6/f12.txt, mode:100644, content:c12]"))))))))))));
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),592,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
        } finally {
            fr.inria.diversify.testamplification.logger.Logger.writeTestFinish(Thread.currentThread());
        }
    }

    @SuppressWarnings(value = "unused")
    private File prepare_f1_1(final Git git) throws IOException {
        return writeTrashFile(F1, "c1");
    }

    private File prepare_f1_2(final Git git) throws Exception {
        final File f1 = prepare_f1_4(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),496,f1);
        f1.delete();
        return f1;
    }

    private File prepare_f1_3(final Git git) throws Exception {
        final File f1 = prepare_f1_7(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),497,f1);
        git.rm().addFilepattern(F1).call();
        return f1;
    }

    private File prepare_f1_4(final Git git) throws Exception {
        final File f1 = prepare_f1_1(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),498,f1);
        git.add().addFilepattern(F1).call();
        return f1;
    }

    private File prepare_f1_5(final Git git) throws Exception {
        final File f1 = prepare_f1_7(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),499,f1);
        f1.delete();
        return f1;
    }

    private File prepare_f1_6(final Git git) throws Exception {
        final File f1 = prepare_f1_3(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),500,f1);
        write(f1, "c1");
        return f1;
    }

    private File prepare_f1_7(final Git git) throws Exception {
        final File f1 = prepare_f1_4(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),501,f1);
        git.commit().setOnly(F1).setMessage(MSG).call();
        return f1;
    }

    private File prepare_f1_8(final Git git) throws Exception {
        final File f1 = prepare_f1_4(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),502,f1);
        write(f1, "c1\'");
        return f1;
    }

    private File prepare_f1_9(final Git git) throws Exception {
        final File f1 = prepare_f1_3(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),503,f1);
        write(f1, "c1\'");
        return f1;
    }

    private File prepare_f1_10(final Git git) throws Exception {
        final File f1 = prepare_f1_9(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),491,f1);
        git.add().addFilepattern(F1).call();
        f1.delete();
        return f1;
    }

    private File prepare_f1_11(final Git git) throws Exception {
        final File f1 = prepare_f1_7(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),492,f1);
        write(f1, "c1\'");
        return f1;
    }

    private File prepare_f1_12(final Git git) throws Exception {
        final File f1 = prepare_f1_13(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),493,f1);
        write(f1, "c1");
        return f1;
    }

    private File prepare_f1_13(final Git git) throws Exception {
        final File f1 = prepare_f1_11(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),494,f1);
        git.add().addFilepattern(F1).call();
        return f1;
    }

    private File prepare_f1_14(final Git git) throws Exception {
        final File f1 = prepare_f1_13(git);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),495,f1);
        write(f1, "c1\'\'");
        return f1;
    }

    private void executeAndCheck_f1_1(final Git git, final int state) throws Exception {
        JGitInternalException exception = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),511,exception);
        try {
            git.commit().setOnly(F1).setMessage(MSG).call();
        } catch (JGitInternalException e) {
            exception = e;
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),510,exception);
        }
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),504,exception);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),505,exception.getMessage().contains(org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),506,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),507,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),508,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),509,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_1_f2_f14(final Git git, final int state) throws Exception {
        JGitInternalException exception = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),533,exception);
        try {
            git.commit().setOnly(F1).setOnly(F2).setMessage(MSG).call();
        } catch (JGitInternalException e) {
            exception = e;
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),532,exception);
        }
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),525,exception);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),526,exception.getMessage().contains(org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),527,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F2));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),528,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),529,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),530,("[f2.txt, mode:100644, content:c2\']" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),531,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_2(final Git git, final int state) throws Exception {
        JGitInternalException exception = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),541,exception);
        try {
            git.commit().setOnly(F1).setMessage(MSG).call();
        } catch (JGitInternalException e) {
            exception = e;
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),540,exception);
        }
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),534,exception);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),535,exception.getMessage().contains("No changes"));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),536,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),537,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),538,("[f1.txt, mode:100644, content:c1]" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),539,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_2_f2_f14(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setOnly(F2).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),542,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),543,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F2));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),544,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),545,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),546,("[f2.txt, mode:100644, content:c2\'\']" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),547,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_3(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),548,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),549,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),550,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),551,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),552,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_4(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),553,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),554,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),555,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),556,("[f1.txt, mode:100644, content:c1]" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),557,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_5(final Git git, final int state) throws Exception {
        executeAndCheck_f1_3(git, state);
    }

    private void executeAndCheck_f1_6(final Git git, final int state) throws Exception {
        JGitInternalException exception = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),565,exception);
        try {
            git.commit().setOnly(F1).setMessage(MSG).call();
        } catch (JGitInternalException e) {
            exception = e;
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),564,exception);
        }
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),558,exception);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),559,exception.getMessage().contains("No changes"));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),560,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),561,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),562,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),563,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_6_f2_14(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setOnly(F2).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),566,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),567,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F2));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),568,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),569,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),570,(("[f1.txt, mode:100644, content:c1]" + "[f2.txt, mode:100644, content:c2\'\']") + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),571,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_7(final Git git, final int state) throws Exception {
        executeAndCheck_f1_2(git, state);
    }

    private void executeAndCheck_f1_7_f2_14(final Git git, final int state) throws Exception {
        executeAndCheck_f1_6_f2_14(git, state);
    }

    private void executeAndCheck_f1_8(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),572,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),573,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),574,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),575,("[f1.txt, mode:100644, content:c1\']" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),576,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_9(final Git git, final int state) throws Exception {
        executeAndCheck_f1_8(git, state);
    }

    private void executeAndCheck_f1_10(final Git git, final int state) throws Exception {
        executeAndCheck_f1_3(git, state);
    }

    private void executeAndCheck_f1_11(final Git git, final int state) throws Exception {
        executeAndCheck_f1_8(git, state);
    }

    private void executeAndCheck_f1_12(final Git git, final int state) throws Exception {
        JGitInternalException exception = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),519,exception);
        try {
            git.commit().setOnly(F1).setMessage(MSG).call();
        } catch (JGitInternalException e) {
            exception = e;
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),518,exception);
        }
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),512,exception);
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),513,exception.getMessage().contains("No changes"));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),514,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),515,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),516,("[f1.txt, mode:100644, content:c1\']" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),517,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void executeAndCheck_f1_12_f2_14(final Git git, final int state) throws Exception {
        executeAndCheck_f1_6_f2_14(git, state);
    }

    private void executeAndCheck_f1_13(final Git git, final int state) throws Exception {
        executeAndCheck_f1_8(git, state);
    }

    private void executeAndCheck_f1_14(final Git git, final int state) throws Exception {
        git.commit().setOnly(F1).setMessage(MSG).call();
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),520,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F1));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),521,org.eclipse.jgit.api.CommitOnlyTest.expected_f3_head(state));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),522,org.eclipse.jgit.api.CommitOnlyTest.getHead(git, org.eclipse.jgit.api.CommitOnlyTest.F3));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),523,("[f1.txt, mode:100644, content:c1\'\']" + (org.eclipse.jgit.api.CommitOnlyTest.expected_f3_idx(state))));
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),524,indexState(org.eclipse.jgit.junit.RepositoryTestCase.CONTENT));
    }

    private void prepare_f3(final Git git, final int state) throws Exception {
        prepare_f3_f2_14(git, state, false);
    }

    private void prepare_f3_f2_14(final Git git, final int state) throws Exception {
        prepare_f3_f2_14(git, state, true);
    }

    private void prepare_f3_f2_14(final Git git, final int state, final boolean include_f2) throws Exception {
        File f2 = null;
        fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),578,f2);
        if (include_f2) {
            f2 = writeTrashFile(F2, "c2");
            fr.inria.diversify.testamplification.logger.Logger.logAssertArgument(Thread.currentThread(),577,f2);
            git.add().addFilepattern(F2).call();
            git.commit().setMessage(MSG).call();
        } 
        if (state >= 1) {
            writeTrashFile(F3, "c3");
            git.add().addFilepattern(F3).call();
        } 
        if (state >= 2)
            git.commit().setMessage(MSG).call();
        
        if (state >= 3)
            git.rm().addFilepattern(F3).call();
        
        if (state == 4) {
            writeTrashFile(F3, "c3\'");
            git.add().addFilepattern(F3).call();
        } 
        if (include_f2) {
            write(f2, "c2\'");
            git.add().addFilepattern(F2).call();
            write(f2, "c2\'\'");
        } 
    }

    private static String expected_f3_head(final int state) {
        switch (state) {
            case 0 :
            case 1 :
                return "";
            case 2 :
            case 3 :
            case 4 :
                return "c3";
        }
        return null;
    }

    private static String expected_f3_idx(final int state) {
        switch (state) {
            case 0 :
            case 3 :
                return "";
            case 1 :
            case 2 :
                return "[f3.txt, mode:100644, content:c3]";
            case 4 :
                return "[f3.txt, mode:100644, content:c3\']";
        }
        return null;
    }

    private static String getHead(final Git git, final String path) throws Exception {
        try {
            final Repository repo = git.getRepository();
            final ObjectId headId = repo.resolve(((org.eclipse.jgit.lib.Constants.HEAD) + "^{commit}"));
            final TreeWalk tw = TreeWalk.forPath(repo, path, new org.eclipse.jgit.revwalk.RevWalk(repo).parseTree(headId));
            return new String(tw.getObjectReader().open(tw.getObjectId(0)).getBytes());
        } catch (Exception e) {
            return "";
        }
    }
}

